import numpy as np
import pymc as pm

from rich.console import Console

from pymc_extras.printing import model_table


def get_text(table) -> str:
    console = Console(width=80)
    with console.capture() as capture:
        console.print(table)
    return capture.get()


def test_model_table():
    with pm.Model(coords={"trial": range(6), "subject": range(20)}) as model:
        x_data = pm.Data("x_data", np.random.normal(size=(6, 20)), dims=("trial", "subject"))
        y_data = pm.Data("y_data", np.random.normal(size=(6, 20)), dims=("trial", "subject"))

        mu = pm.Normal("mu", mu=0, sigma=1)
        sigma = pm.HalfNormal("sigma", sigma=1)
        global_intercept = pm.Normal("global_intercept", mu=0, sigma=1)
        intercept_subject = pm.Normal("intercept_subject", mu=0, sigma=1, shape=(20, 1))
        beta_subject = pm.Normal("beta_subject", mu=mu, sigma=sigma, dims="subject")

        mu_trial = pm.Deterministic(
            "mu_trial",
            global_intercept.squeeze() + intercept_subject + beta_subject * x_data,
            dims=["trial", "subject"],
        )
        noise = pm.Exponential("noise", lam=1)
        y = pm.Normal("y", mu=mu_trial, sigma=noise, observed=y_data, dims=("trial", "subject"))

        pm.Potential("beta_subject_penalty", -pm.math.abs(beta_subject), dims="subject")

    table_txt = get_text(model_table(model))
    expected = """               Variable  Expression                      Dimensions
────────────────────────────────────────────────────────────────────────────────
               x_data =  Data                            trial[6] × subject[20]
               y_data =  Data                            trial[6] × subject[20]

                   mu ~  Normal(0, 1)
                sigma ~  HalfNormal(0, 1)
     global_intercept ~  Normal(0, 1)
    intercept_subject ~  Normal(0, 1)                    [20, 1]
         beta_subject ~  Normal(mu, sigma)               subject[20]
                noise ~  Exponential(f())
                                                         Parameter count = 44

             mu_trial =  f(intercept_subject,            trial[6] × subject[20]
                         beta_subject,
                         global_intercept)

 beta_subject_penalty =  Potential(f(beta_subject))      subject[20]

                    y ~  Normal(mu_trial, noise)         trial[6] × subject[20]
"""
    assert [s.strip() for s in table_txt.splitlines()] == [s.strip() for s in expected.splitlines()]

    table_txt = get_text(model_table(model, split_groups=False))
    expected = """               Variable  Expression                      Dimensions
────────────────────────────────────────────────────────────────────────────────
               x_data =  Data                            trial[6] × subject[20]
               y_data =  Data                            trial[6] × subject[20]
                   mu ~  Normal(0, 1)
                sigma ~  HalfNormal(0, 1)
     global_intercept ~  Normal(0, 1)
    intercept_subject ~  Normal(0, 1)                    [20, 1]
         beta_subject ~  Normal(mu, sigma)               subject[20]
             mu_trial =  f(intercept_subject,            trial[6] × subject[20]
                         beta_subject,
                         global_intercept)
                noise ~  Exponential(f())
                    y ~  Normal(mu_trial, noise)         trial[6] × subject[20]
 beta_subject_penalty =  Potential(f(beta_subject))      subject[20]
                                                         Parameter count = 44
"""
    assert [s.strip() for s in table_txt.splitlines()] == [s.strip() for s in expected.splitlines()]

    table_txt = get_text(
        model_table(model, split_groups=False, truncate_deterministic=30, parameter_count=False)
    )
    expected = """               Variable  Expression                  Dimensions
────────────────────────────────────────────────────────────────────────────
               x_data =  Data                        trial[6] × subject[20]
               y_data =  Data                        trial[6] × subject[20]
                   mu ~  Normal(0, 1)
                sigma ~  HalfNormal(0, 1)
     global_intercept ~  Normal(0, 1)
    intercept_subject ~  Normal(0, 1)                [20, 1]
         beta_subject ~  Normal(mu, sigma)           subject[20]
             mu_trial =  f(intercept_subject, ...)   trial[6] × subject[20]
                noise ~  Exponential(f())
                    y ~  Normal(mu_trial, noise)     trial[6] × subject[20]
 beta_subject_penalty =  Potential(f(beta_subject))  subject[20]
"""
    assert [s.strip() for s in table_txt.splitlines()] == [s.strip() for s in expected.splitlines()]



from itertools import pairwise, product

import numpy as np
import pandas as pd
import pymc as pm
import pytensor
import pytensor.tensor as pt
import pytest
import statsmodels.api as sm

from numpy.testing import assert_allclose, assert_array_less
from pymc.model.transform.optimization import freeze_dims_and_data
from pymc.testing import mock_sample_setup_and_teardown

from pymc_extras.statespace import BayesianVARMAX
from pymc_extras.statespace.utils.constants import SHORT_NAME_TO_LONG
from tests.statespace.shared_fixtures import (  # pylint: disable=unused-import
    rng,
)

mock_sample = pytest.fixture(scope="function")(mock_sample_setup_and_teardown)

floatX = pytensor.config.floatX
ps = [0, 1, 2, 3]
qs = [0, 1, 2, 3]
orders = list(product(ps, qs))[1:]
ids = [f"p={x[0]}, q={x[1]}" for x in orders]


@pytest.fixture(scope="session")
def data():
    df = pd.read_csv(
        "tests/statespace/_data/statsmodels_macrodata_processed.csv",
        index_col=0,
        parse_dates=True,
    ).astype(floatX)
    df.index.freq = df.index.inferred_freq
    return df


@pytest.fixture(scope="session")
def varma_mod(data):
    return BayesianVARMAX(
        endog_names=data.columns,
        order=(2, 0),
        stationary_initialization=True,
        verbose=False,
        measurement_error=True,
    )


@pytest.fixture(scope="session")
def pymc_mod(varma_mod, data):
    with pm.Model(coords=varma_mod.coords) as pymc_mod:
        # x0 = pm.Normal("x0", dims=["state"])
        # P0_diag = pm.Exponential("P0_diag", 1, size=varma_mod.k_states)
        # P0 = pm.Deterministic(
        #     "P0", pt.diag(P0_diag), dims=["state", "state_aux"]
        # )
        state_chol, *_ = pm.LKJCholeskyCov(
            "state_chol", n=varma_mod.k_posdef, eta=1, sd_dist=pm.Exponential.dist(1)
        )
        ar_params = pm.Normal(
            "ar_params", mu=0, sigma=0.1, dims=["observed_state", "lag_ar", "observed_state_aux"]
        )
        state_cov = pm.Deterministic(
            "state_cov", state_chol @ state_chol.T, dims=["shock", "shock_aux"]
        )
        sigma_obs = pm.Exponential("sigma_obs", 1, dims=["observed_state"])

        varma_mod.build_statespace_graph(data=data, save_kalman_filter_outputs_in_idata=True)

    return pymc_mod


@pytest.fixture(scope="session")
def idata(pymc_mod, rng):
    with pymc_mod:
        idata = pm.sample_prior_predictive(draws=10, random_seed=rng)

    return idata


def test_mode_argument():
    # Mode argument should be passed to the parent class
    mod = BayesianVARMAX(endog_names=["y1", "y2"], order=(3, 0), mode="FAST_RUN", verbose=False)
    assert mod.mode == "FAST_RUN"


@pytest.mark.parametrize("order", orders, ids=ids)
@pytest.mark.parametrize("var", ["AR", "MA", "state_cov"])
@pytest.mark.filterwarnings("ignore::statsmodels.tools.sm_exceptions.EstimationWarning")
def test_VARMAX_param_counts_match_statsmodels(data, order, var):
    p, q = order

    mod = BayesianVARMAX(
        endog_names=["realgdp", "realcons", "realinv"], order=(p, q), verbose=False
    )
    sm_var = sm.tsa.VARMAX(data, order=(p, q))

    count = mod.param_counts[var]
    if var == "state_cov":
        # Statsmodels only counts the lower triangle
        count = mod.k_posdef * (mod.k_posdef - 1)
    assert count == sm_var.parameters[var.lower()]


@pytest.mark.parametrize("order", orders, ids=ids)
@pytest.mark.filterwarnings("ignore::statsmodels.tools.sm_exceptions.EstimationWarning")
@pytest.mark.filterwarnings("ignore::FutureWarning")
def test_VARMAX_update_matches_statsmodels(data, order, rng):
    p, q = order

    sm_var = sm.tsa.VARMAX(data, order=(p, q))

    param_counts = [None, *np.cumsum(list(sm_var.parameters.values())).tolist()]
    param_slices = [slice(a, b) for a, b in pairwise(param_counts)]
    param_lists = [trend, ar, ma, reg, state_cov, obs_cov] = [
        sm_var.param_names[idx] for idx in param_slices
    ]
    param_d = {
        k: getattr(np, floatX)(rng.normal(scale=0.1) ** 2)
        for param_list in param_lists
        for k in param_list
    }

    res = sm_var.fit_constrained(param_d)

    mod = BayesianVARMAX(
        endog_names=["realgdp", "realcons", "realinv"],
        order=(p, q),
        verbose=False,
        measurement_error=False,
        stationary_initialization=False,
    )

    ar_shape = (mod.k_endog, mod.p, mod.k_endog)
    ma_shape = (mod.k_endog, mod.q, mod.k_endog)

    with pm.Model() as pm_mod:
        x0 = pm.Deterministic("x0", pt.zeros(mod.k_states, dtype=floatX))
        P0 = pm.Deterministic("P0", pt.eye(mod.k_states, dtype=floatX))
        ma_params = pm.Deterministic(
            "ma_params",
            pt.as_tensor_variable(np.array([param_d[var] for var in ma])).reshape(ma_shape),
        )
        ar_params = pm.Deterministic(
            "ar_params",
            pt.as_tensor_variable(np.array([param_d[var] for var in ar])).reshape(ar_shape),
        )
        state_chol = np.zeros((mod.k_posdef, mod.k_posdef), dtype=floatX)
        state_chol[np.tril_indices(mod.k_posdef)] = np.array([param_d[var] for var in state_cov])
        state_cov = pm.Deterministic("state_cov", pt.as_tensor_variable(state_chol @ state_chol.T))
        mod._insert_random_variables()

        matrices = pm.draw(mod.subbed_ssm)
        matrix_dict = dict(zip(SHORT_NAME_TO_LONG.values(), matrices))

    for matrix in ["transition", "selection", "state_cov", "obs_cov", "design"]:
        assert_allclose(matrix_dict[matrix], sm_var.ssm[matrix])


@pytest.mark.parametrize("filter_output", ["filtered", "predicted", "smoothed"])
def test_all_prior_covariances_are_PSD(filter_output, pymc_mod, rng):
    rv = pymc_mod[f"{filter_output}_covariances"]
    cov_mats = pm.draw(rv, 100, random_seed=rng)
    w, v = np.linalg.eig(cov_mats)
    assert_array_less(0, w, err_msg=f"Smallest eigenvalue: {min(w.ravel())}")


parameters = [
    {"n_steps": 10, "shock_size": None},
    {"n_steps": 10, "shock_size": 1.0},
    {"n_steps": 10, "shock_size": np.array([1.0, 0.0, 0.0])},
    {
        "n_steps": 10,
        "shock_cov": np.array([[1.38, 0.58, -1.84], [0.58, 0.99, -0.82], [-1.84, -0.82, 2.51]]),
    },
    {
        "shock_trajectory": np.r_[
            np.zeros((3, 3), dtype=floatX),
            np.array([[1.0, 0.0, 0.0]]).astype(floatX),
            np.zeros((6, 3), dtype=floatX),
        ]
    },
]

ids = ["from-posterior-cov", "scalar_shock_size", "array_shock_size", "user-cov", "trajectory"]


@pytest.mark.parametrize("parameters", parameters, ids=ids)
@pytest.mark.skipif(floatX == "float32", reason="Impulse covariance not PSD if float32")
def test_impulse_response(parameters, varma_mod, idata, rng):
    irf = varma_mod.impulse_response_function(idata.prior, random_seed=rng, **parameters)

    assert np.isfinite(irf.irf.values).all()


def test_forecast(varma_mod, idata, rng):
    forecast = varma_mod.forecast(idata.prior, periods=10, random_seed=rng)

    assert np.isfinite(forecast.forecast_latent.values).all()
    assert np.isfinite(forecast.forecast_observed.values).all()


def test_varmax_workflow(rng, mock_sample):
    df = pd.read_csv(
        "tests/statespace/_data/statsmodels_macrodata_processed.csv",
        index_col=0,
        parse_dates=True,
    ).astype(floatX)
    df.index.freq = df.index.inferred_freq

    ss_mod = BayesianVARMAX(
        endog_names=df.columns,
        order=(1, 0),
        stationary_initialization=True,
        measurement_error=True,
        verbose=False,
    )

    with pm.Model(coords=ss_mod.coords) as m:
        state_cov_diag = pm.Exponential("state_cov_diag", 1, dims=["shock"])
        pm.Deterministic("state_cov", pt.diag(state_cov_diag), dims=["shock", "shock_aux"])
        pm.Normal("ar_params", sigma=0.1, dims=["observed_state", "lag_ar", "observed_state_aux"])
        pm.Exponential("sigma_obs", 1, dims=["observed_state"])

        ss_mod.build_statespace_graph(df)

        idata = pm.sample()

    post = ss_mod.sample_conditional_posterior(idata, mvn_method="svd")
    assert "filtered_posterior" in post
    assert "smoothed_posterior" in post
    assert "predicted_posterior" in post

    forecast = ss_mod.forecast(idata, periods=10, random_seed=rng)
    assert "forecast_latent" in forecast
    assert "forecast_observed" in forecast
    assert np.isfinite(forecast.forecast_latent.values).all()
    assert np.isfinite(forecast.forecast_observed.values).all()

    irf = ss_mod.impulse_response_function(idata, n_steps=10, random_seed=rng)
    assert "irf" in irf
    assert np.isfinite(irf.irf.values).all()


class TestVARMAXWithExogenous:
    def test_create_varmax_with_exogenous_list_of_names(self, data):
        mod = BayesianVARMAX(
            endog_names=["realgdp", "realcons", "realinv"],
            order=(1, 0),
            exog_state_names=["foo", "bar"],
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )
        assert mod.k_exog == 2
        assert mod.exog_state_names == ["foo", "bar"]
        assert mod.data_names == ("exogenous_data",)
        assert mod.param_dims["beta_exog"] == ("observed_state", "exogenous")
        assert mod.coords["exogenous"] == ("foo", "bar")
        assert mod.param_info["beta_exog"]["shape"] == (mod.k_endog, 2)
        assert mod.param_info["beta_exog"]["dims"] == ("observed_state", "exogenous")

    def test_create_varmax_with_exogenous_both_defined_correctly(self, data):
        mod = BayesianVARMAX(
            endog_names=["realgdp", "realcons", "realinv"],
            order=(1, 0),
            exog_state_names=["a", "b"],
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )
        assert mod.k_exog == 2
        assert mod.exog_state_names == ["a", "b"]
        assert mod.data_names == ("exogenous_data",)
        assert mod.param_dims["beta_exog"] == ("observed_state", "exogenous")
        assert mod.coords["exogenous"] == ("a", "b")
        assert mod.param_info["beta_exog"]["shape"] == (mod.k_endog, 2)
        assert mod.param_info["beta_exog"]["dims"] == ("observed_state", "exogenous")

    def test_create_varmax_with_exogenous_exog_names_dict(self, data):
        exog_state_names = {"observed_0": ["a", "b"], "observed_1": ["c"], "observed_2": []}
        mod = BayesianVARMAX(
            endog_names=["observed_0", "observed_1", "observed_2"],
            order=(1, 0),
            exog_state_names=exog_state_names,
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )
        assert mod.k_exog == {"observed_0": 2, "observed_1": 1, "observed_2": 0}
        assert mod.exog_state_names == exog_state_names
        assert mod.data_names == (
            "observed_0_exogenous_data",
            "observed_1_exogenous_data",
            "observed_2_exogenous_data",
        )
        assert mod.param_dims["beta_observed_0"] == ("exogenous_observed_0",)
        assert mod.param_dims["beta_observed_1"] == ("exogenous_observed_1",)
        assert (
            "beta_observed_2" not in mod.param_dims
            or mod.param_info.get("beta_observed_2") is None
            or mod.param_info.get("beta_observed_2", {}).get("shape", (0,))[0] == 0
        )

        assert mod.coords["exogenous_observed_0"] == ("a", "b")
        assert mod.coords["exogenous_observed_1"] == ("c",)
        assert "exogenous_observed_2" in mod.coords and mod.coords["exogenous_observed_2"] == ()

        assert mod.param_info["beta_observed_0"]["shape"] == (2,)
        assert mod.param_info["beta_observed_0"]["dims"] == ("exogenous_observed_0",)
        assert mod.param_info["beta_observed_1"]["shape"] == (1,)
        assert mod.param_info["beta_observed_1"]["dims"] == ("exogenous_observed_1",)

    def test_create_varmax_with_exogenous_dict_converts_to_list(self, data):
        exog_state_names = {
            "observed_0": ["a", "b"],
            "observed_1": ["a", "b"],
            "observed_2": ["a", "b"],
        }
        mod = BayesianVARMAX(
            endog_names=["observed_0", "observed_1", "observed_2"],
            order=(1, 0),
            exog_state_names=exog_state_names,
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
        )

        assert mod.k_exog == 2
        assert mod.exog_state_names == ["a", "b"]
        assert mod.data_names == ("exogenous_data",)
        assert mod.param_dims["beta_exog"] == ("observed_state", "exogenous")
        assert mod.coords["exogenous"] == ("a", "b")
        assert mod.param_info["beta_exog"]["shape"] == (mod.k_endog, 2)
        assert mod.param_info["beta_exog"]["dims"] == ("observed_state", "exogenous")

    def _build_varmax(self, df, exog_state_names, exog_data):
        endog_names = df.columns.values.tolist()

        mod = BayesianVARMAX(
            endog_names=endog_names,
            order=(1, 0),
            exog_state_names=exog_state_names,
            verbose=False,
            measurement_error=False,
            stationary_initialization=False,
            mode="JAX",
        )

        with pm.Model(coords=mod.coords) as m:
            for var_name, data in exog_data.items():
                pm.Data(var_name, data, dims=mod.data_info[var_name]["dims"])

            x0 = pm.Deterministic("x0", pt.zeros(mod.k_states), dims=mod.param_dims["x0"])
            P0_diag = pm.Exponential("P0_diag", 1.0, dims=mod.param_dims["P0"][0])
            P0 = pm.Deterministic("P0", pt.diag(P0_diag), dims=mod.param_dims["P0"])

            ar_params = pm.Normal("ar_params", mu=0, sigma=1, dims=mod.param_dims["ar_params"])
            state_cov_diag = pm.Exponential(
                "state_cov_diag", 1.0, dims=mod.param_dims["state_cov"][0]
            )
            state_cov = pm.Deterministic(
                "state_cov", pt.diag(state_cov_diag), dims=mod.param_dims["state_cov"]
            )

            # Exogenous priors
            if isinstance(mod.exog_state_names, list):
                beta_exog = pm.Normal("beta_exog", mu=0, sigma=1, dims=mod.param_dims["beta_exog"])
            elif isinstance(mod.exog_state_names, dict):
                for name in mod.exog_state_names:
                    if mod.exog_state_names.get(name):
                        pm.Normal(
                            f"beta_{name}", mu=0, sigma=1, dims=mod.param_dims[f"beta_{name}"]
                        )

            mod.build_statespace_graph(data=df)

        return mod, m

    @pytest.mark.parametrize(
        "exog_state_names",
        [
            (["foo", "bar"]),
            ({"y1": ["a", "b"], "y2": ["c"]}),
        ],
        ids=["exog_state_names_list", "exog_state_names_dict"],
    )
    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_varmax_with_exog(self, rng, exog_state_names):
        endog_names = ["y1", "y2", "y3"]
        n_obs = 50
        time_idx = pd.date_range(start="2020-01-01", periods=n_obs, freq="D")

        y = rng.normal(size=(n_obs, len(endog_names)))
        df = pd.DataFrame(y, columns=endog_names, index=time_idx).astype(floatX)

        if isinstance(exog_state_names, dict):
            exog_data = {
                f"{name}_exogenous_data": pd.DataFrame(
                    rng.normal(size=(n_obs, len(exog_names))).astype(floatX),
                    columns=exog_names,
                    index=time_idx,
                )
                for name, exog_names in exog_state_names.items()
            }
        else:
            exog_data = {
                "exogenous_data": pd.DataFrame(
                    rng.normal(size=(n_obs, len(exog_state_names))).astype(floatX),
                    columns=exog_state_names,
                    index=time_idx,
                )
            }

        mod, m = self._build_varmax(df, exog_state_names, exog_data)

        with freeze_dims_and_data(m):
            prior = pm.sample_prior_predictive(
                draws=10, random_seed=rng, compile_kwargs={"mode": "JAX"}
            )

        prior_cond = mod.sample_conditional_prior(prior, mvn_method="eigh")
        beta_dot_data = prior_cond.filtered_prior_observed.values - prior_cond.filtered_prior.values

        if isinstance(exog_state_names, list):
            beta = prior.prior.beta_exog
            assert beta.shape == (1, 10, 3, 2)

            np.testing.assert_allclose(
                beta_dot_data,
                np.einsum("tx,...sx->...ts", exog_data["exogenous_data"].values, beta),
                atol=1e-2,
            )

        elif isinstance(exog_state_names, dict):
            assert prior.prior.beta_y1.shape == (1, 10, 2)
            assert prior.prior.beta_y2.shape == (1, 10, 1)

            obs_intercept = [
                np.einsum("tx,...x->...t", exog_data[f"{name}_exogenous_data"].values, beta)
                for name, beta in zip(["y1", "y2"], [prior.prior.beta_y1, prior.prior.beta_y2])
            ]

            # y3 has no exogenous variables
            obs_intercept.append(np.zeros_like(obs_intercept[0]))

            np.testing.assert_allclose(beta_dot_data, np.stack(obs_intercept, axis=-1), atol=1e-2)

    @pytest.mark.filterwarnings("ignore::UserWarning")
    def test_forecast_with_exog(self, rng):
        endog_names = ["y1", "y2", "y3"]
        n_obs = 50
        time_idx = pd.date_range(start="2020-01-01", periods=n_obs, freq="D")

        y = rng.normal(size=(n_obs, len(endog_names)))
        df = pd.DataFrame(y, columns=endog_names, index=time_idx).astype(floatX)

        mod, m = self._build_varmax(
            df,
            exog_state_names=["exogenous_0", "exogenous_1"],
            exog_data={
                "exogenous_data": pd.DataFrame(
                    rng.normal(size=(n_obs, 2)).astype(floatX),
                    columns=["exogenous_0", "exogenous_1"],
                    index=time_idx,
                )
            },
        )

        assert mod._needs_exog_data

        with freeze_dims_and_data(m):
            prior = pm.sample_prior_predictive(
                draws=10, random_seed=rng, compile_kwargs={"mode": "JAX"}
            )

        with pytest.raises(
            ValueError,
            match="This model was fit using exogenous data. Forecasting cannot be performed "
            "without providing scenario data",
        ):
            mod.forecast(prior.prior, periods=10, random_seed=rng)

        forecast = mod.forecast(
            prior.prior,
            periods=10,
            random_seed=rng,
            scenario={
                "exogenous_data": pd.DataFrame(
                    rng.normal(size=(10, 2)).astype(floatX),
                    columns=["exogenous_0", "exogenous_1"],
                    index=pd.date_range(start=df.index[-1], periods=10, freq="D"),
                )
            },
        )

        assert np.isfinite(forecast.forecast_latent.values).all()
        assert np.isfinite(forecast.forecast_observed.values).all()



import numpy as np
import pytensor

from numpy.testing import assert_allclose
from pytensor import config
from pytensor.graph.traversal import explicit_graph_inputs
from scipy import linalg

from pymc_extras.statespace.models import structural as st
from pymc_extras.statespace.models.structural.utils import _frequency_transition_block
from tests.statespace.models.structural.conftest import _assert_basic_coords_correct
from tests.statespace.test_utilities import assert_pattern_repeats, simulate_from_numpy_model

ATOL = 1e-8 if config.floatX.endswith("64") else 1e-4
RTOL = 0 if config.floatX.endswith("64") else 1e-6


cycle_test_vals = zip([None, None, 3, 5, 10], [False, True, True, False, False])


def test_cycle_component_deterministic(rng):
    cycle = st.Cycle(name="cycle", cycle_length=12, estimate_cycle_length=False, innovations=False)
    params = {"params_cycle": np.array([1.0, 1.0], dtype=config.floatX)}
    x, y = simulate_from_numpy_model(cycle, rng, params, steps=12 * 12)

    assert_pattern_repeats(y, 12, atol=ATOL, rtol=RTOL)


def test_cycle_component_with_dampening(rng):
    cycle = st.Cycle(
        name="cycle", cycle_length=12, estimate_cycle_length=False, innovations=False, dampen=True
    )
    params = {
        "params_cycle": np.array([10.0, 10.0], dtype=config.floatX),
        "dampening_factor_cycle": 0.75,
    }
    x, y = simulate_from_numpy_model(cycle, rng, params, steps=100)

    # check that cycle dampens to zero over time
    assert_allclose(y[-1], 0.0, atol=ATOL, rtol=RTOL)


def test_cycle_component_with_innovations_and_cycle_length(rng):
    cycle = st.Cycle(name="cycle", estimate_cycle_length=True, innovations=True, dampen=True)
    params = {
        "params_cycle": np.array([1.0, 1.0], dtype=config.floatX),
        "length_cycle": 12.0,
        "dampening_factor_cycle": 0.95,
        "sigma_cycle": 1.0,
    }
    x, y = simulate_from_numpy_model(cycle, rng, params)

    cycle.build(verbose=False)
    assert cycle.coords["state_cycle"] == ("Cos_cycle", "Sin_cycle")


def test_cycle_multivariate_deterministic(rng):
    """Test multivariate cycle component with deterministic cycles."""
    cycle = st.Cycle(
        name="cycle",
        cycle_length=12,
        estimate_cycle_length=False,
        innovations=False,
        observed_state_names=["data_1", "data_2", "data_3"],
    )
    params = {"params_cycle": np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]], dtype=config.floatX)}
    x, y = simulate_from_numpy_model(cycle, rng, params, steps=12 * 12)

    # Check that each variable has a cyclical pattern with the expected period
    for i in range(3):
        assert_pattern_repeats(y[:, i], 12, atol=ATOL, rtol=RTOL)

    # Check that the cycles have different amplitudes (different initial states)
    assert np.std(y[:, 0]) > 0
    assert np.std(y[:, 1]) > 0
    assert np.std(y[:, 2]) > 0
    # The second and third variables should have larger amplitudes due to larger initial states
    assert np.std(y[:, 1]) > np.std(y[:, 0])
    assert np.std(y[:, 2]) > np.std(y[:, 0])

    # check design, transition, selection matrices
    Z, T, R = pytensor.function(
        [],
        [cycle.ssm["design"], cycle.ssm["transition"], cycle.ssm["selection"]],
        mode="FAST_COMPILE",
    )()

    # each block is [1, 0] for design
    expected_Z = np.zeros((3, 6))
    expected_Z[0, 0] = 1.0
    expected_Z[1, 2] = 1.0
    expected_Z[2, 4] = 1.0
    np.testing.assert_allclose(Z, expected_Z)

    # each block is 2x2 frequency transition matrix for given cycle length (12 here)
    block = _frequency_transition_block(12, 1).eval()
    expected_T = np.zeros((6, 6))
    for i in range(3):
        expected_T[2 * i : 2 * i + 2, 2 * i : 2 * i + 2] = block
    np.testing.assert_allclose(T, expected_T)

    # each block is 2x2 identity for selection
    expected_R = np.zeros((6, 6))
    for i in range(3):
        expected_R[2 * i : 2 * i + 2, 2 * i : 2 * i + 2] = np.eye(2)
    np.testing.assert_allclose(R, expected_R)


def test_multivariate_cycle_with_shared(rng):
    cycle = st.Cycle(
        name="cycle",
        cycle_length=12,
        estimate_cycle_length=False,
        innovations=False,
        observed_state_names=["data_1", "data_2", "data_3"],
        share_states=True,
    )

    assert cycle.state_names == (
        "Cos_cycle[shared]",
        "Sin_cycle[shared]",
    )
    assert cycle.shock_names == ()
    assert cycle.param_names == ("params_cycle",)

    params = {"params_cycle": np.array([1.0, 2.0], dtype=config.floatX)}
    x, y = simulate_from_numpy_model(cycle, rng, params, steps=12 * 12)

    np.testing.assert_allclose(y[:, 0], y[:, 1], atol=ATOL, rtol=RTOL)
    np.testing.assert_allclose(y[:, 0], y[:, 2], atol=ATOL, rtol=RTOL)


def test_cycle_multivariate_with_dampening(rng):
    """Test multivariate cycle component with dampening."""
    cycle = st.Cycle(
        name="cycle",
        cycle_length=12,
        estimate_cycle_length=False,
        innovations=False,
        dampen=True,
        observed_state_names=["data_1", "data_2", "data_3"],
    )
    params = {
        "params_cycle": np.array([[10.0, 10.0], [20.0, 20.0], [30.0, 30.0]], dtype=config.floatX),
        "dampening_factor_cycle": 0.75,
    }
    x, y = simulate_from_numpy_model(cycle, rng, params, steps=100)

    # Check that all cycles dampen to zero over time
    for i in range(3):
        assert_allclose(y[-1, i], 0.0, atol=ATOL, rtol=RTOL)

    # Check that the dampening pattern is consistent across variables
    # The variables should dampen at the same rate but with different initial amplitudes
    for i in range(1, 3):
        # The ratio of final to initial values should be similar across variables
        ratio_0 = abs(y[-1, 0] / y[0, 0]) if y[0, 0] != 0 else 0
        ratio_i = abs(y[-1, i] / y[0, i]) if y[0, i] != 0 else 0
        assert_allclose(ratio_0, ratio_i, atol=1e-2, rtol=1e-2)


def test_cycle_multivariate_with_innovations_and_cycle_length(rng):
    """Test multivariate cycle component with innovations and estimated cycle length."""
    cycle = st.Cycle(
        name="cycle",
        estimate_cycle_length=True,
        innovations=True,
        dampen=True,
        observed_state_names=["data_1", "data_2", "data_3"],
    )
    params = {
        "params_cycle": np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]], dtype=config.floatX),
        "length_cycle": 12.0,
        "dampening_factor_cycle": 0.95,
        "sigma_cycle": np.array([0.5, 1.0, 1.5]),  # different innov variances per var
    }
    x, y = simulate_from_numpy_model(cycle, rng, params)

    mod = cycle.build(verbose=False)
    _assert_basic_coords_correct(mod)

    assert cycle.coords["state_cycle"] == ("Cos_cycle", "Sin_cycle")
    assert cycle.coords["endog_cycle"] == (
        "data_1",
        "data_2",
        "data_3",
    )

    assert cycle.k_endog == 3
    assert cycle.k_states == 6  # 2 states per variable
    assert cycle.k_posdef == 6  # 2 innovations per variable

    # Check that the data has the expected shape
    assert y.shape[1] == 3  # 3 variables

    # Check that each variable shows some variation (due to innovations)
    for i in range(3):
        assert np.std(y[:, i]) > 0

    # check design, transition, selection & state_cov matrices
    Z, T, R, Q = pytensor.function(
        [
            cycle._name_to_variable["length_cycle"],
            cycle._name_to_variable["dampening_factor_cycle"],
            cycle._name_to_variable["sigma_cycle"],
        ],
        [
            cycle.ssm["design"],
            cycle.ssm["transition"],
            cycle.ssm["selection"],
            cycle.ssm["state_cov"],
        ],
        mode="FAST_COMPILE",
    )(params["length_cycle"], params["dampening_factor_cycle"], params["sigma_cycle"])

    # each block is [1, 0] for design
    expected_Z = np.zeros((3, 6))
    expected_Z[0, 0] = 1.0
    expected_Z[1, 2] = 1.0
    expected_Z[2, 4] = 1.0
    np.testing.assert_allclose(Z, expected_Z)

    # each block is 2x2 frequency transition matrix for given cycle length (12 here),
    # scaled by dampening factor
    block = _frequency_transition_block(12, 1).eval() * params["dampening_factor_cycle"]
    expected_T = np.zeros((6, 6))
    for i in range(3):
        expected_T[2 * i : 2 * i + 2, 2 * i : 2 * i + 2] = block
    np.testing.assert_allclose(T, expected_T)

    # each block is 2x2 identity for selection
    expected_R = np.zeros((6, 6))
    for i in range(3):
        expected_R[2 * i : 2 * i + 2, 2 * i : 2 * i + 2] = np.eye(2)
    np.testing.assert_allclose(R, expected_R)

    # each block is sigma^2 * I_2 for state_cov
    sigmas = params["sigma_cycle"]
    expected_Q = np.zeros((6, 6))
    for i in range(3):
        expected_Q[2 * i : 2 * i + 2, 2 * i : 2 * i + 2] = np.eye(2) * sigmas[i] ** 2
    np.testing.assert_allclose(Q, expected_Q)


def test_add_multivariate_cycle_components_with_different_observed():
    """
    Test adding two multivariate Cycles with different observed_state_names.
    Ensures that combining two multivariate Cycles with different observed state names
    results in the correct block-diagonal state space matrices and state naming.
    """
    cycle1 = st.Cycle(
        name="cycle1",
        cycle_length=12,
        estimate_cycle_length=False,
        innovations=False,
        observed_state_names=["a1", "a2"],
    )
    cycle2 = st.Cycle(
        name="cycle2",
        cycle_length=6,
        estimate_cycle_length=False,
        innovations=False,
        observed_state_names=["b1", "b2"],
    )
    mod = (cycle1 + cycle2).build(verbose=False)

    # check dimensions
    assert mod.k_endog == 4
    assert mod.k_states == 8
    assert mod.k_posdef == 2 * mod.k_endog  # 2 innovations per variable

    # check state names and coords
    expected_state_names = (
        "Cos_cycle1[a1]",
        "Sin_cycle1[a1]",
        "Cos_cycle1[a2]",
        "Sin_cycle1[a2]",
        "Cos_cycle2[b1]",
        "Sin_cycle2[b1]",
        "Cos_cycle2[b2]",
        "Sin_cycle2[b2]",
    )
    assert mod.state_names == expected_state_names

    assert mod.coords["state_cycle1"] == ("Cos_cycle1", "Sin_cycle1")
    assert mod.coords["state_cycle2"] == ("Cos_cycle2", "Sin_cycle2")
    assert mod.coords["endog_cycle1"] == (
        "a1",
        "a2",
    )
    assert mod.coords["endog_cycle2"] == (
        "b1",
        "b2",
    )

    # evaluate design, transition, selection matrices
    Z, T, R = pytensor.function(
        [], [mod.ssm["design"], mod.ssm["transition"], mod.ssm["selection"]], mode="FAST_COMPILE"
    )()

    # design: each row selects first state of its block
    expected_Z = np.zeros((4, 8))
    expected_Z[0, 0] = 1.0  # "a1" -> Cos_cycle1[a1]
    expected_Z[1, 2] = 1.0  # "a2" -> Cos_cycle1[a2]
    expected_Z[2, 4] = 1.0  # "b1" -> Cos_cycle2[b1]
    expected_Z[3, 6] = 1.0  # "b2" -> Cos_cycle2[b2]
    assert_allclose(Z, expected_Z)

    # transition: block diagonal, each block is 2x2 frequency transition matrix
    block1 = _frequency_transition_block(12, 1).eval()
    block2 = _frequency_transition_block(6, 1).eval()
    expected_T = np.zeros((8, 8))
    for i in range(2):
        expected_T[2 * i : 2 * i + 2, 2 * i : 2 * i + 2] = block1
    for i in range(2):
        expected_T[4 + 2 * i : 4 + 2 * i + 2, 4 + 2 * i : 4 + 2 * i + 2] = block2
    assert_allclose(T, expected_T)

    # selection: block diagonal, each block is 2x2 identity
    expected_R = np.zeros((8, 8))
    for i in range(4):
        expected_R[2 * i : 2 * i + 2, 2 * i : 2 * i + 2] = np.eye(2)
    assert_allclose(R, expected_R)


def test_add_multivariate_shared_and_not_shared():
    cycle_shared = st.Cycle(
        name="shared_cycle",
        cycle_length=12,
        estimate_cycle_length=False,
        innovations=True,
        observed_state_names=["gdp", "inflation", "unemployment"],
        share_states=True,
    )
    cycle_individual = st.Cycle(
        name="individual_cycle",
        estimate_cycle_length=True,
        innovations=False,
        observed_state_names=["gdp", "inflation", "unemployment"],
        dampen=True,
    )
    mod = (cycle_shared + cycle_individual).build(verbose=False)

    assert mod.k_endog == 3
    assert mod.k_states == 2 + 3 * 2
    assert mod.k_posdef == 2 + 3 * 2

    expected_states = (
        "Cos_shared_cycle[shared]",
        "Sin_shared_cycle[shared]",
        "Cos_individual_cycle[gdp]",
        "Sin_individual_cycle[gdp]",
        "Cos_individual_cycle[inflation]",
        "Sin_individual_cycle[inflation]",
        "Cos_individual_cycle[unemployment]",
        "Sin_individual_cycle[unemployment]",
    )

    assert mod.state_names == expected_states
    assert mod.shock_names == expected_states[:2]

    assert mod.param_names == (
        "params_shared_cycle",
        "sigma_shared_cycle",
        "params_individual_cycle",
        "length_individual_cycle",
        "dampening_factor_individual_cycle",
        "P0",
    )

    assert "endog_shared_cycle" not in mod.coords
    assert mod.coords["state_shared_cycle"] == ("Cos_shared_cycle", "Sin_shared_cycle")
    assert mod.coords["state_individual_cycle"] == ("Cos_individual_cycle", "Sin_individual_cycle")
    assert mod.coords["endog_individual_cycle"] == (
        "gdp",
        "inflation",
        "unemployment",
    )

    assert mod._param_info["params_shared_cycle"].dims == ("state_shared_cycle",)
    assert mod._param_info["params_shared_cycle"].shape == (2,)

    assert mod._param_info["sigma_shared_cycle"].dims is None
    assert mod._param_info["sigma_shared_cycle"].shape == ()

    assert mod._param_info["params_individual_cycle"].dims == (
        "endog_individual_cycle",
        "state_individual_cycle",
    )
    assert mod._param_info["params_individual_cycle"].shape == (3, 2)

    params = {
        "length_individual_cycle": 12.0,
        "dampening_factor_individual_cycle": 0.95,
    }
    outputs = [mod.ssm["transition"], mod.ssm["design"], mod.ssm["selection"]]
    T, Z, R = pytensor.function(
        list(explicit_graph_inputs(outputs)),
        outputs,
        mode="FAST_COMPILE",
    )(**params)

    lamb = 2 * np.pi / 12  # dampening factor for individual cycle
    transition_block = np.array(
        [[np.cos(lamb), np.sin(lamb)], [-np.sin(lamb), np.cos(lamb)]], dtype=config.floatX
    )
    T_expected = linalg.block_diag(transition_block, *[0.95 * transition_block] * 3)
    np.testing.assert_allclose(T, T_expected)

    np.testing.assert_allclose(
        Z, np.array([[1, 0, 1, 0, 0, 0, 0, 0], [1, 0, 0, 0, 1, 0, 0, 0], [1, 0, 0, 0, 0, 0, 1, 0]])
    )

    np.testing.assert_allclose(R, np.eye(8, dtype=config.floatX))



#   Copyright 2022 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging

from importlib.metadata import version

from pymc_extras import gp, statespace, utils
from pymc_extras.distributions import *
from pymc_extras.inference import find_MAP, fit, fit_laplace, fit_pathfinder
from pymc_extras.model.marginal.marginal_model import (
    MarginalModel,
    marginalize,
    recover_marginals,
)
from pymc_extras.model.model_api import as_model

_log = logging.getLogger("pmx")

if not logging.root.handlers:
    _log.setLevel(logging.INFO)
    if len(_log.handlers) == 0:
        handler = logging.StreamHandler()
        _log.addHandler(handler)


__version__ = version("pymc-extras")



import numpy as np
import pytensor

from pytensor.graph.traversal import explicit_graph_inputs

from pymc_extras.statespace.models import structural as st
from tests.statespace.models.structural.conftest import _assert_basic_coords_correct


def test_measurement_error(rng):
    mod = st.MeasurementError("obs") + st.LevelTrend(order=2)
    mod = mod.build(verbose=False)

    assert "sigma_obs" in mod.param_names


def test_measurement_error_multiple_observed():
    mod = st.MeasurementError("obs", observed_state_names=["data_1", "data_2"])
    assert mod.k_endog == 2
    assert mod.coords["endog_obs"] == (
        "data_1",
        "data_2",
    )
    assert mod.param_info["sigma_obs"].dims == ("endog_obs",)


def test_measurement_error_share_states():
    mod = st.MeasurementError("obs", observed_state_names=["data_1", "data_2"], share_states=True)
    mod = mod.build(verbose=False)
    _assert_basic_coords_correct(mod)

    assert mod.k_endog == 2
    assert mod.param_names == (
        "sigma_obs",
        "P0",
    )
    assert "endog_obs" not in mod.coords

    # Check that the parameter is shared across the observed states
    assert mod._param_info["sigma_obs"].shape == ()

    outputs = mod.ssm["obs_cov"]

    H = pytensor.function(list(explicit_graph_inputs([outputs])), outputs)(sigma_obs=np.array(0.5))
    np.testing.assert_allclose(H, np.diag([0.5, 0.5]) ** 2)


def test_measurement_error_shared_and_not_shared():
    shared = st.MeasurementError(
        "error_shared", observed_state_names=["data_1", "data_2"], share_states=True
    )
    individual = st.MeasurementError("error_individual", observed_state_names=["data_1", "data_2"])
    mod = (shared + individual).build(verbose=False)

    assert mod.k_endog == 2
    assert mod.param_names == (
        "sigma_error_shared",
        "sigma_error_individual",
        "P0",
    )
    assert mod.coords["endog_error_individual"] == (
        "data_1",
        "data_2",
    )

    assert mod._param_info["sigma_error_shared"].shape == ()
    assert mod._param_info["sigma_error_individual"].shape == (2,)

    outputs = mod.ssm["obs_cov"]

    H = pytensor.function(list(explicit_graph_inputs([outputs])), outputs)(
        sigma_error_shared=np.array(0.5), sigma_error_individual=np.array([0.1, 0.9])
    )
    np.testing.assert_allclose(H, np.diag([0.5, 0.5]) ** 2 + np.diag([0.1, 0.9]) ** 2)


def test_build_with_measurement_error_subset():
    ll = st.LevelTrend(order=2, observed_state_names=["data_1", "data_2", "data_3"])
    me = st.MeasurementError("obs", observed_state_names=["data_1", "data_3"])
    mod = (ll + me).build()

    H = mod.ssm["obs_cov"]
    assert H.type.shape == (3, 3)
    np.testing.assert_allclose(
        H.eval({"sigma_obs": [1.0, 3.0]}),
        np.array([[1.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 9.0]]),
    )



#   Copyright 2023 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pymc as pm
import pytensor.tensor as pt
import pytest
import scipy

from numpy import dtype
from xarray.core.utils import Frozen

jax = pytest.importorskip("jax")
pytest.importorskip("blackjax")

from pymc_extras.inference.smc.sampling import (
    arviz_from_particles,
    blackjax_particles_from_pymc_population,
    get_jaxified_loglikelihood,
    get_jaxified_logprior,
    sample_smc_blackjax,
)


def two_gaussians_model():
    n = 4
    mu1 = np.ones(n) * 0.5
    mu2 = -mu1

    stdev = 0.1
    sigma = np.power(stdev, 2) * np.eye(n)
    isigma = np.linalg.inv(sigma)
    dsigma = np.linalg.det(sigma)

    w1 = stdev
    w2 = 1 - stdev

    def two_gaussians(x):
        """
        Mixture of gaussians likelihood
        """
        log_like1 = (
            -0.5 * n * pt.log(2 * np.pi)
            - 0.5 * pt.log(dsigma)
            - 0.5 * (x - mu1).T.dot(isigma).dot(x - mu1)
        )
        log_like2 = (
            -0.5 * n * pt.log(2 * np.pi)
            - 0.5 * pt.log(dsigma)
            - 0.5 * (x - mu2).T.dot(isigma).dot(x - mu2)
        )
        return pt.log(w1 * pt.exp(log_like1) + w2 * pt.exp(log_like2))

    with pm.Model() as m:
        X = pm.Uniform("X", lower=-2, upper=2.0, shape=n)
        llk = pm.Potential("muh", two_gaussians(X))

    return m, mu1


def fast_model():
    with pm.Model() as m:
        x = pm.Normal("x", 0, 1)
        y = pm.Normal("y", x, 1, observed=0)
    return m


@pytest.mark.parametrize(
    "kernel, check_for_integration_steps, inner_kernel_params",
    [
        ("HMC", True, {"step_size": 0.1, "integration_steps": 11}),
        ("NUTS", False, {"step_size": 0.1}),
    ],
)
def test_sample_smc_blackjax(kernel, check_for_integration_steps, inner_kernel_params):
    """
    When running the two gaussians model
    with BlackJax SMC, we sample them correctly,
    the shape of a posterior variable is (1, particles, dimension)
    and the inference_data has the right attributes.

    """
    model, muref = two_gaussians_model()
    iterations_to_diagnose = 2
    n_particles = 1000
    with model:
        inference_data = sample_smc_blackjax(
            n_particles=n_particles,
            kernel=kernel,
            inner_kernel_params=inner_kernel_params,
            iterations_to_diagnose=iterations_to_diagnose,
        )

    x = inference_data.posterior["X"]

    assert x.to_numpy().shape == (1, n_particles, 4)
    mu1d = np.abs(x).mean(axis=0).mean(axis=0)
    np.testing.assert_allclose(muref, mu1d, rtol=0.0, atol=0.03)

    for attribute, value in [
        ("particles", n_particles),
        ("step_size", 0.1),
        ("num_mcmc_steps", 10),
        ("iterations_to_diagnose", iterations_to_diagnose),
        ("sampler", f"Blackjax SMC with {kernel} kernel"),
    ]:
        assert inference_data.posterior.attrs[attribute] == value

    for diagnostic in ["lambda_evolution", "log_likelihood_increments"]:
        assert inference_data.posterior.attrs[diagnostic].shape == (iterations_to_diagnose,)

    for diagnostic in ["ancestors_evolution", "weights_evolution"]:
        assert inference_data.posterior.attrs[diagnostic].shape == (
            iterations_to_diagnose,
            n_particles,
        )

    for attribute in ["running_time_seconds", "iterations"]:
        assert attribute in inference_data.posterior.attrs

    if check_for_integration_steps:
        assert inference_data.posterior.attrs["integration_steps"] == 11


def test_blackjax_particles_from_pymc_population_univariate():
    model = fast_model()
    population = {"x": np.array([2, 3, 4])}
    blackjax_particles = blackjax_particles_from_pymc_population(model, population)
    jax.tree.map(np.testing.assert_allclose, blackjax_particles, [np.array([[2], [3], [4]])])


def test_blackjax_particles_from_pymc_population_multivariate():
    with pm.Model() as model:
        x = pm.Normal("x", 0, 1)
        z = pm.Normal("z", 0, 1)
        y = pm.Normal("y", x + z, 1, observed=0)

    population = {"x": np.array([0.34614613, 1.09163261, -0.44526825]), "z": np.array([1, 2, 3])}
    blackjax_particles = blackjax_particles_from_pymc_population(model, population)
    jax.tree.map(
        np.testing.assert_allclose,
        blackjax_particles,
        [np.array([[0.34614613], [1.09163261], [-0.44526825]]), np.array([[1], [2], [3]])],
    )


def simple_multivariable_model():
    """
    A simple model that has a multivariate variable,
    a has more than one variable (multivariable)
    """
    with pm.Model() as model:
        x = pm.Normal("x", 0, 1, shape=2)
        z = pm.Normal("z", 0, 1)
        y = pm.Normal("y", z, 1, observed=0)
    return model


def test_blackjax_particles_from_pymc_population_multivariable():
    model = simple_multivariable_model()
    population = {"x": np.array([[2, 3], [5, 6], [7, 9]]), "z": np.array([11, 12, 13])}
    blackjax_particles = blackjax_particles_from_pymc_population(model, population)

    jax.tree.map(
        np.testing.assert_allclose,
        blackjax_particles,
        [np.array([[2, 3], [5, 6], [7, 9]]), np.array([[11], [12], [13]])],
    )


def test_arviz_from_particles():
    model = simple_multivariable_model()
    particles = [np.array([[2, 3], [5, 6], [7, 9]]), np.array([[11], [12], [13]])]
    with model:
        inference_data = arviz_from_particles(model, particles)

    assert inference_data.posterior.sizes == Frozen({"chain": 1, "draw": 3, "x_dim_0": 2})
    assert inference_data.posterior.data_vars.dtypes == Frozen(
        {"x": dtype("float64"), "z": dtype("float64")}
    )


def test_get_jaxified_logprior():
    """
    Given a model with a Normal prior
    for a RV, the jaxified logprior
    indeed calculates that number,
    and can be jax.vmap'ed
    """
    logprior = get_jaxified_logprior(fast_model())
    for point in [-0.5, 0.0, 0.5]:
        jax.tree.map(
            np.testing.assert_allclose,
            jax.vmap(logprior)([np.array([point])]),
            np.log(scipy.stats.norm(0, 1).pdf(point)),
        )


def test_get_jaxified_loglikelihood():
    """
    Given a model with a Normal Likelihood, a single observation
    0 and std=1, the only free parameter of that function is the mean.
    When computing the logliklikelihood
    Then the function can be jax.vmap'ed, and the calculation matches the likelihood.
    """
    loglikelihood = get_jaxified_loglikelihood(fast_model())
    for point in [-0.5, 0.0, 0.5]:
        jax.tree.map(
            np.testing.assert_allclose,
            jax.vmap(loglikelihood)([np.array([point])]),
            np.log(scipy.stats.norm(point, 1).pdf(0)),
        )



from collections.abc import Callable
from functools import cache

import numpy as np
import pytensor
import pytensor.tensor as pt
import pytest

from numpy.testing import assert_allclose, assert_array_less

from pymc_extras.statespace.filters import (
    KalmanSmoother,
    SquareRootFilter,
    StandardFilter,
    UnivariateFilter,
)
from pymc_extras.statespace.filters.kalman_filter import BaseFilter
from tests.statespace.shared_fixtures import (  # pylint: disable=unused-import
    rng,
)
from tests.statespace.test_utilities import (
    get_expected_shape,
    get_sm_state_from_output_name,
    initialize_filter,
    make_test_inputs,
    nile_test_test_helper,
)

floatX = pytensor.config.floatX

# TODO: These are pretty loose because of all the stabilizing of covariance matrices that is done inside the kalman
#  filters. When that is improved, this should be tightened.
ATOL = 1e-6 if floatX.endswith("64") else 1e-3
RTOL = 1e-6 if floatX.endswith("64") else 1e-3


@cache
def get_filter_function(filter_name: str) -> Callable:
    """
    Compile and return a filter function given its name, caching the result to make tests as fast as possible
    """
    match filter_name:
        case "StandardFilter":
            filter_inout = initialize_filter(StandardFilter())
        case "CholeskyFilter":
            filter_inout = initialize_filter(SquareRootFilter())
        case "UnivariateFilter":
            filter_inout = initialize_filter(UnivariateFilter())
        case _:
            raise ValueError(f"Unknown filter name: {filter_name}")

    filter_func = pytensor.function(*filter_inout, on_unused_input="ignore")
    return filter_func


filter_names = [
    "StandardFilter",
    "CholeskyFilter",
    "UnivariateFilter",
]

output_names = [
    "filtered_states",
    "predicted_states",
    "smoothed_states",
    "filtered_covs",
    "predicted_covs",
    "smoothed_covs",
    "log_likelihood",
    "ll_obs",
]


def test_base_class_update_raises():
    filter = BaseFilter()
    inputs = [None] * 7
    with pytest.raises(NotImplementedError):
        filter.update(*inputs)


@pytest.mark.parametrize("filter_name", filter_names)
def test_output_shapes_one_state_one_observed(filter_name, rng):
    p, m, r, n = 1, 1, 1, 10
    inputs = make_test_inputs(p, m, r, n, rng)
    outputs = get_filter_function(filter_name)(*inputs)

    for output_idx, name in enumerate(output_names):
        expected_output = get_expected_shape(name, p, m, r, n)
        assert (
            outputs[output_idx].shape == expected_output
        ), f"Shape of {name} does not match expected"


@pytest.mark.parametrize("filter_name", filter_names)
def test_output_shapes_when_all_states_are_stochastic(filter_name, rng):
    p, m, r, n = 1, 2, 2, 10
    inputs = make_test_inputs(p, m, r, n, rng)

    outputs = get_filter_function(filter_name)(*inputs)
    for output_idx, name in enumerate(output_names):
        expected_output = get_expected_shape(name, p, m, r, n)
        assert (
            outputs[output_idx].shape == expected_output
        ), f"Shape of {name} does not match expected"


@pytest.mark.parametrize("filter_name", filter_names)
def test_output_shapes_when_some_states_are_deterministic(filter_name, rng):
    p, m, r, n = 1, 5, 2, 10
    inputs = make_test_inputs(p, m, r, n, rng)

    outputs = get_filter_function(filter_name)(*inputs)
    for output_idx, name in enumerate(output_names):
        expected_output = get_expected_shape(name, p, m, r, n)
        assert (
            outputs[output_idx].shape == expected_output
        ), f"Shape of {name} does not match expected"


@pytest.fixture
def f_standard_nd():
    ksmoother = KalmanSmoother()
    data = pt.tensor(name="data", dtype=floatX, shape=(None, None))
    a0 = pt.vector(name="a0", dtype=floatX)
    P0 = pt.matrix(name="P0", dtype=floatX)
    c = pt.vector(name="c", dtype=floatX)
    d = pt.vector(name="d", dtype=floatX)
    Q = pt.tensor(name="Q", dtype=floatX, shape=(None, None, None))
    H = pt.tensor(name="H", dtype=floatX, shape=(None, None, None))
    T = pt.tensor(name="T", dtype=floatX, shape=(None, None, None))
    R = pt.tensor(name="R", dtype=floatX, shape=(None, None, None))
    Z = pt.tensor(name="Z", dtype=floatX, shape=(None, None, None))

    inputs = [data, a0, P0, c, d, T, Z, R, H, Q]

    (
        filtered_states,
        predicted_states,
        observed_states,
        filtered_covs,
        predicted_covs,
        observed_covs,
        ll_obs,
    ) = StandardFilter().build_graph(*inputs)

    smoothed_states, smoothed_covs = ksmoother.build_graph(T, R, Q, filtered_states, filtered_covs)

    outputs = [
        filtered_states,
        predicted_states,
        smoothed_states,
        filtered_covs,
        predicted_covs,
        smoothed_covs,
        ll_obs.sum(),
        ll_obs,
    ]

    f_standard = pytensor.function(inputs, outputs)

    return f_standard


def test_output_shapes_with_time_varying_matrices(f_standard_nd, rng):
    p, m, r, n = 1, 5, 2, 10
    data, a0, P0, c, d, T, Z, R, H, Q = make_test_inputs(p, m, r, n, rng)
    T = np.concatenate([np.expand_dims(T, 0)] * n, axis=0)
    Z = np.concatenate([np.expand_dims(Z, 0)] * n, axis=0)
    R = np.concatenate([np.expand_dims(R, 0)] * n, axis=0)
    H = np.concatenate([np.expand_dims(H, 0)] * n, axis=0)
    Q = np.concatenate([np.expand_dims(Q, 0)] * n, axis=0)

    outputs = f_standard_nd(data, a0, P0, c, d, T, Z, R, H, Q)

    for output_idx, name in enumerate(output_names):
        expected_output = get_expected_shape(name, p, m, r, n)
        assert (
            outputs[output_idx].shape == expected_output
        ), f"Shape of {name} does not match expected"


@pytest.mark.parametrize("filter_name", filter_names)
def test_output_with_deterministic_observation_equation(filter_name, rng):
    p, m, r, n = 1, 5, 1, 10
    inputs = make_test_inputs(p, m, r, n, rng)

    outputs = get_filter_function(filter_name)(*inputs)

    for output_idx, name in enumerate(output_names):
        expected_output = get_expected_shape(name, p, m, r, n)
        assert (
            outputs[output_idx].shape == expected_output
        ), f"Shape of {name} does not match expected"


@pytest.mark.parametrize("filter_name", filter_names)
def test_output_with_multiple_observed(filter_name, rng):
    p, m, r, n = 5, 5, 1, 10
    inputs = make_test_inputs(p, m, r, n, rng)

    outputs = get_filter_function(filter_name)(*inputs)
    for output_idx, name in enumerate(output_names):
        expected_output = get_expected_shape(name, p, m, r, n)
        assert (
            outputs[output_idx].shape == expected_output
        ), f"Shape of {name} does not match expected"


@pytest.mark.parametrize("filter_name", filter_names)
@pytest.mark.parametrize("p", [1, 5], ids=["univariate (p=1)", "multivariate (p=5)"])
def test_missing_data(filter_name, p, rng):
    m, r, n = 5, 1, 10
    inputs = make_test_inputs(p, m, r, n, rng, missing_data=1)

    outputs = get_filter_function(filter_name)(*inputs)
    for output_idx, name in enumerate(output_names):
        expected_output = get_expected_shape(name, p, m, r, n)
        assert (
            outputs[output_idx].shape == expected_output
        ), f"Shape of {name} does not match expected"


@pytest.mark.parametrize("filter_name", filter_names)
@pytest.mark.parametrize("output_idx", [(0, 2), (3, 5)], ids=["smoothed_states", "smoothed_covs"])
def test_last_smoother_is_last_filtered(filter_name, output_idx, rng):
    p, m, r, n = 1, 5, 1, 10
    inputs = make_test_inputs(p, m, r, n, rng)
    outputs = get_filter_function(filter_name)(*inputs)

    filtered = outputs[output_idx[0]]
    smoothed = outputs[output_idx[1]]

    assert_allclose(filtered[-1], smoothed[-1])


@pytest.mark.parametrize("filter_name", filter_names)
@pytest.mark.parametrize("n_missing", [0, 5], ids=["n_missing=0", "n_missing=5"])
@pytest.mark.skipif(floatX == "float32", reason="Tests are too sensitive for float32")
def test_filters_match_statsmodel_output(filter_name, n_missing, rng):
    fit_sm_mod, [data, a0, P0, c, d, T, Z, R, H, Q] = nile_test_test_helper(rng, n_missing)
    if filter_name == "CholeskyFilter":
        P0 = np.linalg.cholesky(P0)
    inputs = [data, a0, P0, c, d, T, Z, R, H, Q]
    outputs = get_filter_function(filter_name)(*inputs)

    for output_idx, name in enumerate(output_names):
        ref_val = get_sm_state_from_output_name(fit_sm_mod, name)
        val_to_test = outputs[output_idx].squeeze()

        if name == "smoothed_covs":
            # TODO: The smoothed covariance matrices have large errors (1e-2) ONLY in the first few states -- no idea why.
            assert_allclose(
                val_to_test[5:],
                ref_val[5:],
                atol=ATOL,
                rtol=RTOL,
                err_msg=f"{name} does not match statsmodels",
            )
        elif name.startswith("predicted"):
            # statsmodels doesn't throw away the T+1 forecast in the predicted states like we do
            assert_allclose(
                val_to_test,
                ref_val[:-1],
                atol=ATOL,
                rtol=RTOL,
                err_msg=f"{name} does not match statsmodels",
            )
        else:
            # Need atol = 1e-7 for smoother tests to pass
            assert_allclose(
                val_to_test,
                ref_val,
                atol=ATOL,
                rtol=RTOL,
                err_msg=f"{name} does not match statsmodels",
            )


@pytest.mark.parametrize("filter_name", filter_names)
@pytest.mark.parametrize("n_missing", [0, 5], ids=["n_missing=0", "n_missing=5"])
@pytest.mark.parametrize("obs_noise", [True, False])
def test_all_covariance_matrices_are_PSD(filter_name, n_missing, obs_noise, rng):
    if (floatX == "float32") & (filter_name == "UnivariateFilter"):
        # TODO: These tests all pass locally for me with float32 but they fail on the CI, so i'm just disabling them.
        pytest.skip("Univariate filter not stable at half precision without measurement error")

    fit_sm_mod, [data, a0, P0, c, d, T, Z, R, H, Q] = nile_test_test_helper(rng, n_missing)
    if filter_name == "CholeskyFilter":
        P0 = np.linalg.cholesky(P0)

    H *= int(obs_noise)
    inputs = [data, a0, P0, c, d, T, Z, R, H, Q]
    outputs = get_filter_function(filter_name)(*inputs)

    for output_idx, name in zip([3, 4, 5], output_names[3:-2]):
        cov_stack = outputs[output_idx]
        w, v = np.linalg.eig(cov_stack)

        assert_array_less(0, w, err_msg=f"Smallest eigenvalue of {name}: {min(w.ravel())}")
        assert_allclose(
            cov_stack,
            np.swapaxes(cov_stack, -2, -1),
            rtol=RTOL,
            atol=ATOL,
            err_msg=f"{name} is not symmetrical",
        )


@pytest.mark.parametrize(
    "filter",
    [StandardFilter, SquareRootFilter],
    ids=["standard", "cholesky"],
)
def test_kalman_filter_jax(filter):
    pytest.importorskip("jax")
    from pymc.sampling.jax import get_jaxified_graph

    # TODO: Add UnivariateFilter to test; need to figure out the broadcasting issue when 2nd data dim is defined

    p, m, r, n = 1, 5, 1, 10
    inputs, outputs = initialize_filter(filter(), p=p, m=m, r=r, n=n)
    inputs_np = make_test_inputs(p, m, r, n, rng)

    f_jax = get_jaxified_graph(inputs, outputs)
    f_pt = pytensor.function(inputs, outputs, mode="FAST_COMPILE")

    jax_outputs = f_jax(*inputs_np)
    pt_outputs = f_pt(*inputs_np)

    for name, jax_res, pt_res in zip(output_names, jax_outputs, pt_outputs):
        assert_allclose(jax_res, pt_res, atol=ATOL, rtol=RTOL, err_msg=f"{name} failed!")



import warnings

import numpy as np
import pandas as pd
import pymc as pm
import pytensor
import pytensor.tensor as pt

from pymc import ImputationWarning, modelcontext
from pytensor.tensor.sharedvar import TensorSharedVariable

from pymc_extras.statespace.utils.constants import (
    MISSING_FILL,
    OBS_STATE_DIM,
    TIME_DIM,
)

NO_TIME_INDEX_WARNING = (
    "No time index found on the supplied data. A simple range index will be automatically "
    "generated."
)
NO_FREQ_INFO_WARNING = "No frequency was specific on the data's DateTimeIndex."


def get_data_dims(data):
    if not isinstance(data, pt.TensorVariable | TensorSharedVariable):
        return

    data_name = getattr(data, "name", None)
    if not data_name:
        return

    pm_mod = modelcontext(None)
    data_dims = pm_mod.named_vars_to_dims.get(data_name, None)
    return data_dims


def _validate_data_shape(data_shape, n_obs, obs_coords=None, check_col_names=False, col_names=None):
    if col_names is None:
        col_names = []

    if len(data_shape) != 2:
        raise ValueError("Data must be a 2d matrix")

    if data_shape[-1] != n_obs:
        raise ValueError(
            f"Shape of data does not match model output. Expected {n_obs} columns, "
            f"found {data_shape[-1]}."
        )

    if check_col_names:
        missing_cols = set(obs_coords) - set(col_names)
        if len(missing_cols) > 0:
            raise ValueError(
                "Columns of DataFrame provided as data do not match state names. The following states were"
                f"not found: {', '.join(missing_cols)}. This may result in unexpected results in complex"
                f"statespace models"
            )


def preprocess_tensor_data(data, n_obs, obs_coords=None):
    data_shape = data.shape.eval()
    _validate_data_shape(data_shape, n_obs, obs_coords)
    if obs_coords is not None:
        warnings.warn(NO_TIME_INDEX_WARNING)
    index = np.arange(data_shape[0], dtype="int")

    return data.eval(), index


def preprocess_numpy_data(data, n_obs, obs_coords=None):
    _validate_data_shape(data.shape, n_obs, obs_coords)
    if obs_coords is not None:
        warnings.warn(NO_TIME_INDEX_WARNING)

    index = np.arange(data.shape[0], dtype="int")

    return data, index


def preprocess_pandas_data(data, n_obs, obs_coords=None, check_column_names=False):
    if isinstance(data, pd.Series):
        if data.name is None:
            data.name = "data"
        data = data.to_frame()

    col_names = data.columns
    _validate_data_shape(data.shape, n_obs, obs_coords, check_column_names, col_names)

    if isinstance(data.index, pd.DatetimeIndex):
        if data.index.freq is None:
            warnings.warn(NO_FREQ_INFO_WARNING)
            data.index.freq = data.index.inferred_freq

        index = data.index
        return data.values, index

    elif isinstance(data.index, pd.RangeIndex):
        if obs_coords is not None:
            warnings.warn(NO_TIME_INDEX_WARNING)
        return preprocess_numpy_data(data.values, n_obs, obs_coords)

    elif isinstance(data.index, pd.MultiIndex):
        if obs_coords is not None:
            warnings.warn(NO_TIME_INDEX_WARNING)

        raise NotImplementedError("MultiIndex panel data is not currently supported.")

    else:
        if obs_coords is not None:
            warnings.warn(NO_TIME_INDEX_WARNING)

        index = data.index
        if not np.issubdtype(index.dtype, np.integer):
            raise IndexError("Provided index is not an integer index.")

        index_diff = index.to_series().diff().dropna().values
        if not (index_diff == 1).all():
            raise IndexError("Provided index is not monotonic increasing.")

        return preprocess_numpy_data(data.values, n_obs, obs_coords)


def add_data_to_active_model(values, index, data_dims=None):
    pymc_mod = modelcontext(None)
    if data_dims is None:
        data_dims = [TIME_DIM, OBS_STATE_DIM]
    time_dim = data_dims[0]

    if time_dim not in pymc_mod.coords:
        pymc_mod.add_coord(time_dim, index)
    else:
        found_time = pymc_mod.coords[time_dim]
        if found_time is None:
            pymc_mod.coords.update({time_dim: index})
        elif not np.array_equal(found_time, tuple(index)):
            raise ValueError(
                "Provided data has a different time index than the model. Please ensure that the time values "
                "set on coords matches that of the exogenous data."
            )

    # If the data has just one column, we need to specify the shape as (None, 1), or else the JAX backend will
    # raise a broadcasting error.
    if values.shape[-1] == 1 or values.ndim == 1:
        data_shape = (None, 1)
    else:
        data_shape = (None, values.shape[-1])

    data = pm.Data("data", values, dims=data_dims, shape=data_shape)

    return data


def mask_missing_values_in_data(values, missing_fill_value=None):
    if missing_fill_value is None:
        missing_fill_value = MISSING_FILL

    masked_values = np.ma.masked_invalid(values)
    filled_values = masked_values.filled(missing_fill_value)
    nan_mask = masked_values.mask

    if np.any(nan_mask):
        if np.any(values == missing_fill_value):
            raise ValueError(
                f"Provided data contains the value {missing_fill_value}, which is used as a missing value marker. "
                f"Please manually change the missing_fill_value to avoid this collision."
            )

        impute_message = (
            "Provided data contains missing values and"
            " will be automatically imputed as hidden states"
            " during Kalman filtering."
        )

        warnings.warn(impute_message, ImputationWarning)

    return filled_values, nan_mask


def register_data_with_pymc(
    data, n_obs, obs_coords, register_data=True, missing_fill_value=None, data_dims=None
):
    if isinstance(data, pt.TensorVariable | TensorSharedVariable):
        values, index = preprocess_tensor_data(data, n_obs, obs_coords)
    elif isinstance(data, np.ndarray):
        values, index = preprocess_numpy_data(data, n_obs, obs_coords)
    elif isinstance(data, pd.DataFrame | pd.Series):
        values, index = preprocess_pandas_data(data, n_obs, obs_coords)
    else:
        raise ValueError("Data should be one of pytensor tensor, numpy array, or pandas dataframe")

    data, nan_mask = mask_missing_values_in_data(values, missing_fill_value)

    if register_data:
        data = add_data_to_active_model(data, index, data_dims)
    else:
        data = pytensor.shared(data, name="data")
    return data, nan_mask



import numpy as np
import pymc as pm
import pytensor.tensor as pt
import pytest

from pymc_extras.inference.laplace_approx.find_map import (
    find_MAP,
    get_nearest_psd,
)
from pymc_extras.inference.laplace_approx.scipy_interface import (
    GradientBackend,
    scipy_optimize_funcs_from_loss,
    set_optimizer_function_defaults,
)


@pytest.fixture(scope="session")
def rng():
    seed = sum(map(ord, "test_fit_map"))
    return np.random.default_rng(seed)


def test_get_nearest_psd_returns_psd(rng):
    # Matrix with negative eigenvalues
    A = np.array([[2, -3], [-3, 2]])
    psd = get_nearest_psd(A)

    # Should be symmetric
    np.testing.assert_allclose(psd, psd.T)

    # All eigenvalues should be >= 0
    eigvals = np.linalg.eigvalsh(psd)
    assert np.all(eigvals >= -1e-12), "All eigenvalues should be non-negative"


def test_get_nearest_psd_given_psd_input(rng):
    L = rng.normal(size=(2, 2))
    A = L @ L.T
    psd = get_nearest_psd(A)

    # Given PSD input, should return the same matrix
    assert np.allclose(psd, A)


def test_set_optimizer_function_defaults_warns_and_prefers_hessp(caplog):
    # "trust-ncg" uses_grad=True, uses_hess=True, uses_hessp=True
    method = "trust-ncg"
    with caplog.at_level("WARNING"):
        use_grad, use_hess, use_hessp = set_optimizer_function_defaults(method, True, True, True)

    message = caplog.messages[0]
    assert message.startswith('Both "use_hess" and "use_hessp" are set to True')

    assert use_grad
    assert not use_hess
    assert use_hessp


def test_set_optimizer_function_defaults_infers_hess_and_hessp():
    # "trust-ncg" uses_grad=True, uses_hess=True, uses_hessp=True
    method = "trust-ncg"

    # If only use_hessp is set, use_hess should be False but use_grad should be inferred as True
    use_grad, use_hess, use_hessp = set_optimizer_function_defaults(method, None, None, True)
    assert use_grad
    assert not use_hess
    assert use_hessp

    # Only use_hess is set
    use_grad, use_hess, use_hessp = set_optimizer_function_defaults(method, None, True, None)
    assert use_hess
    assert not use_hessp


def test_set_optimizer_function_defaults_defaults():
    # "trust-ncg" uses_grad=True, uses_hess=True, uses_hessp=True
    method = "trust-ncg"
    use_grad, use_hess, use_hessp = set_optimizer_function_defaults(method, None, None, None)
    assert use_grad
    assert not use_hess
    assert use_hessp


@pytest.mark.parametrize("gradient_backend", ["jax", "pytensor"], ids=str)
def test_jax_functions_from_graph(gradient_backend: GradientBackend):
    pytest.importorskip("jax")

    x = pt.tensor("x", shape=(2,))

    def compute_z(x):
        z1 = x[0] ** 2 + 2
        z2 = x[0] * x[1] + 3
        return z1, z2

    z = pt.stack(compute_z(x))
    f_fused, f_hessp = scipy_optimize_funcs_from_loss(
        loss=z.sum(),
        inputs=[x],
        initial_point_dict={"x": np.array([1.0, 2.0])},
        use_grad=True,
        use_hess=True,
        use_hessp=True,
        gradient_backend=gradient_backend,
        compile_kwargs=dict(mode="JAX"),
    )

    x_val = np.array([1.0, 2.0])
    expected_z = sum(compute_z(x_val))

    z_jax, grad_val, hess_val = f_fused(x_val)
    np.testing.assert_allclose(z_jax, expected_z)
    np.testing.assert_allclose(grad_val.squeeze(), np.array([2 * x_val[0] + x_val[1], x_val[0]]))

    hess_val = np.array(hess_val)
    np.testing.assert_allclose(hess_val.squeeze(), np.array([[2, 1], [1, 0]]))

    hessp_val = np.array(f_hessp(x_val, np.array([1.0, 0.0])))
    np.testing.assert_allclose(hessp_val.squeeze(), np.array([2, 1]))


@pytest.mark.parametrize(
    "method, use_grad, use_hess, use_hessp",
    [
        (
            "Newton-CG",
            True,
            True,
            False,
        ),
        ("Newton-CG", True, False, True),
        ("BFGS", True, False, False),
        ("L-BFGS-B", True, False, False),
    ],
)
@pytest.mark.parametrize(
    "backend, gradient_backend, include_transformed, compute_hessian",
    [("jax", "jax", True, True), ("jax", "pytensor", False, False)],
    ids=str,
)
def test_find_MAP(
    method,
    use_grad,
    use_hess,
    use_hessp,
    backend,
    gradient_backend: GradientBackend,
    include_transformed,
    compute_hessian,
    rng,
):
    pytest.importorskip("jax")

    with pm.Model() as m:
        mu = pm.Normal("mu")
        sigma = pm.Exponential("sigma", 1)
        pm.Normal("y_hat", mu=mu, sigma=sigma, observed=rng.normal(loc=3, scale=1.5, size=10))

        idata = find_MAP(
            method=method,
            use_grad=use_grad,
            use_hess=use_hess,
            use_hessp=use_hessp,
            progressbar=False,
            gradient_backend=gradient_backend,
            include_transformed=include_transformed,
            compile_kwargs={"mode": backend.upper()},
            maxiter=5,
            compute_hessian=compute_hessian,
        )

    assert hasattr(idata, "posterior")
    assert hasattr(idata, "fit")
    assert hasattr(idata, "optimizer_result")
    assert hasattr(idata, "observed_data")

    posterior = idata.posterior.squeeze(["chain", "draw"])
    assert "mu" in posterior and "sigma" in posterior
    assert posterior["mu"].shape == ()
    assert posterior["sigma"].shape == ()

    if include_transformed:
        assert hasattr(idata, "unconstrained_posterior")
        unconstrained_posterior = idata.unconstrained_posterior.squeeze(["chain", "draw"])
        assert "sigma_log__" in unconstrained_posterior
        assert unconstrained_posterior["sigma_log__"].shape == ()
    else:
        assert not hasattr(idata, "unconstrained_posterior")

    assert ("covariance_matrix" in idata.fit) == compute_hessian


def test_find_map_outside_model_context():
    """
    Test that find_MAP can be called outside of a model context.
    """
    with pm.Model() as m:
        mu = pm.Normal("mu", 0, 1)
        sigma = pm.Exponential("sigma", 1)
        y_hat = pm.Normal("y_hat", mu=mu, sigma=sigma, observed=np.random.normal(size=10))

    idata = find_MAP(model=m, method="L-BFGS-B", use_grad=True, progressbar=False)

    assert hasattr(idata, "posterior")
    assert hasattr(idata, "fit")
    assert hasattr(idata, "optimizer_result")


@pytest.mark.parametrize(
    "backend, gradient_backend",
    [("jax", "jax")],
    ids=str,
)
def test_map_shared_variables(backend, gradient_backend: GradientBackend):
    pytest.importorskip("jax")

    with pm.Model() as m:
        data = pm.Data("data", np.random.normal(loc=3, scale=1.5, size=10))
        mu = pm.Normal("mu")
        sigma = pm.Exponential("sigma", 1)
        y_hat = pm.Normal("y_hat", mu=mu, sigma=sigma, observed=data)

        idata = find_MAP(
            method="L-BFGS-B",
            use_grad=True,
            use_hess=False,
            use_hessp=False,
            progressbar=False,
            gradient_backend=gradient_backend,
            compile_kwargs={"mode": backend.upper()},
        )

    assert hasattr(idata, "posterior")
    assert hasattr(idata, "unconstrained_posterior")
    assert hasattr(idata, "fit")
    assert hasattr(idata, "optimizer_result")
    assert hasattr(idata, "observed_data")
    assert hasattr(idata, "constant_data")

    posterior = idata.posterior.squeeze(["chain", "draw"])
    unconstrained_posterior = idata.unconstrained_posterior.squeeze(["chain", "draw"])

    assert "mu" in posterior and "sigma" in posterior
    assert posterior["mu"].shape == ()
    assert posterior["sigma"].shape == ()

    assert "sigma_log__" in unconstrained_posterior
    assert unconstrained_posterior["sigma_log__"].shape == ()


@pytest.mark.parametrize(
    "method, use_grad, use_hess, use_hessp",
    [
        ("Newton-CG", True, True, False),
        ("Newton-CG", True, False, True),
    ],
)
@pytest.mark.parametrize(
    "backend, gradient_backend",
    [("jax", "pytensor")],
    ids=str,
)
def test_find_MAP_basinhopping(
    method, use_grad, use_hess, use_hessp, backend, gradient_backend, rng
):
    pytest.importorskip("jax")

    with pm.Model() as m:
        mu = pm.Normal("mu")
        sigma = pm.Exponential("sigma", 1)
        pm.Normal("y_hat", mu=mu, sigma=sigma, observed=rng.normal(loc=3, scale=1.5, size=10))

        idata = find_MAP(
            method="basinhopping",
            use_grad=use_grad,
            use_hess=use_hess,
            use_hessp=use_hessp,
            progressbar=False,
            gradient_backend=gradient_backend,
            compile_kwargs={"mode": backend.upper()},
            minimizer_kwargs=dict(method=method),
            niter=1,
        )

    assert hasattr(idata, "posterior")
    assert hasattr(idata, "unconstrained_posterior")

    posterior = idata.posterior.squeeze(["chain", "draw"])
    unconstrained_posterior = idata.unconstrained_posterior.squeeze(["chain", "draw"])
    assert "mu" in posterior
    assert posterior["mu"].shape == ()

    assert "sigma_log__" in unconstrained_posterior
    assert unconstrained_posterior["sigma_log__"].shape == ()


def test_find_MAP_with_coords():
    with pm.Model(coords={"group": [1, 2, 3, 4, 5]}) as m:
        mu_loc = pm.Normal("mu_loc", 0, 1)
        mu_scale = pm.HalfNormal("mu_scale", 1)

        mu = pm.Normal("mu", mu_loc, mu_scale, dims=["group"])
        sigma = pm.HalfNormal("sigma", 1, dims=["group"])

        obs = pm.Normal("obs", mu=mu, sigma=sigma, observed=np.random.normal(size=(10, 5)))

        idata = find_MAP(progressbar=False, method="L-BFGS-B")

    assert hasattr(idata, "posterior")
    assert hasattr(idata, "unconstrained_posterior")
    assert hasattr(idata, "fit")

    posterior = idata.posterior.squeeze(["chain", "draw"])
    unconstrained_posterior = idata.unconstrained_posterior.squeeze(["chain", "draw"])

    assert (
        "mu_loc" in posterior
        and "mu_scale" in posterior
        and "mu" in posterior
        and "sigma" in posterior
    )
    assert "mu_scale_log__" in unconstrained_posterior and "sigma_log__" in unconstrained_posterior

    assert posterior["mu_loc"].shape == ()
    assert posterior["mu_scale"].shape == ()
    assert posterior["mu"].shape == (5,)
    assert posterior["sigma"].shape == (5,)

    assert unconstrained_posterior["mu_scale_log__"].shape == ()
    assert unconstrained_posterior["sigma_log__"].shape == (5,)


def test_map_nonscalar_rv_without_dims():
    with pm.Model(coords={"test": ["A", "B", "C"]}) as model:
        x_loc = pm.Normal("x_loc", mu=0, sigma=1, dims=["test"])
        x = pm.Normal("x", mu=x_loc, sigma=1, shape=(2, 3))
        y = pm.Normal("y", mu=x, sigma=1, observed=np.random.randn(10, 2, 3))

        idata = find_MAP(method="L-BFGS-B", progressbar=False)

    assert idata.posterior["x"].shape == (1, 1, 2, 3)
    assert all(f"x_dim_{i}" in idata.posterior.coords for i in range(2))

    assert idata.fit.rows.values.tolist() == [
        "x_loc[A]",
        "x_loc[B]",
        "x_loc[C]",
        "x[0,0]",
        "x[0,1]",
        "x[0,2]",
        "x[1,0]",
        "x[1,1]",
        "x[1,2]",
    ]



#   Copyright 2022 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import re
import sys

import numpy as np
import pymc as pm
import pytensor.tensor as pt
import pytest

import pymc_extras as pmx


def eight_schools_model() -> pm.Model:
    J = 8
    y = np.array([28.0, 8.0, -3.0, 7.0, -1.0, 1.0, 18.0, 12.0])
    sigma = np.array([15.0, 10.0, 16.0, 11.0, 9.0, 11.0, 10.0, 18.0])

    with pm.Model() as model:
        mu = pm.Normal("mu", mu=0.0, sigma=10.0)
        tau = pm.HalfCauchy("tau", 5.0)

        theta = pm.Normal("theta", mu=0, sigma=1, shape=J)
        obs = pm.Normal("obs", mu=mu + tau * theta, sigma=sigma, shape=J, observed=y)

    return model


@pytest.fixture
def reference_idata():
    model = eight_schools_model()
    with model:
        idata = pmx.fit(
            method="pathfinder",
            num_paths=10,
            jitter=12.0,
            random_seed=41,
            inference_backend="pymc",
        )
    return idata


def unstable_lbfgs_update_mask_model() -> pm.Model:
    # data and model from: https://github.com/pymc-devs/pymc-extras/issues/445
    # this scenario made LBFGS struggle leading to a lot of rejected iterations, (result.nit being moderate, but only history.count <= 1).
    # this scenario is used to test that the LBFGS history manager is rejecting iterations as expected and PF can run to completion.

    # fmt: off
    inp = np.array([0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 2, 0, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 2, 0, 1, 0, 0, 0, 0, 1, 1, 1, 2, 0, 1, 2, 1, 0, 1, 0, 1, 0, 1, 0])

    res = np.array([[0,0,1,0,0],[0,0,1,0,0],[0,0,1,0,0],[0,1,0,0,0],[0,0,1,0,0],[0,1,0,0,0],[0,0,0,1,0],[0,0,0,1,0],[1,0,0,0,0],[0,1,0,0,0],[0,0,1,0,0],[1,0,0,0,0],[0,0,1,0,0],[0,1,0,0,0],[0,0,0,1,0],[0,0,1,0,0],[0,0,1,0,0],[0,0,0,1,0],[0,0,0,1,0],[0,1,0,0,0],[0,0,0,1,0],[0,0,1,0,0],[0,1,0,0,0],[1,0,0,0,0],[1,0,0,0,0],[0,0,1,0,0],[0,0,1,0,0],[0,0,1,0,0],[0,0,1,0,0],[0,1,0,0,0],[0,1,0,0,0],[0,0,0,1,0],[0,0,1,0,0],[0,1,0,0,0],[0,0,1,0,0],[0,0,0,1,0],[0,0,1,0,0],[1,0,0,0,0],[1,0,0,0,0],[0,0,1,0,0],[0,0,0,1,0],[0,0,0,1,0],[1,0,0,0,0],[1,0,0,0,0],[0,1,0,0,0],[1,0,0,0,0],[0,0,1,0,0],[0,0,1,0,0],[1,0,0,0,0],[0,0,0,1,0]])
    # fmt: on

    n_ordered = res.shape[1]
    coords = {
        "obs": np.arange(len(inp)),
        "inp": np.arange(max(inp) + 1),
        "outp": np.arange(res.shape[1]),
    }
    with pm.Model(coords=coords) as mdl:
        mu = pm.Normal("intercept", sigma=3.5)[None]

        offset = pm.Normal(
            "offset", dims=("inp"), transform=pm.distributions.transforms.ZeroSumTransform([0])
        )

        scale = 3.5 * pm.HalfStudentT("scale", nu=5)
        mu += (scale * offset)[inp]

        phi_delta = pm.Dirichlet("phi_diffs", [1.0] * (n_ordered - 1))
        phi = pt.concatenate([[0], pt.cumsum(phi_delta)])
        s_mu = pm.Normal(
            "stereotype_intercept",
            size=n_ordered,
            transform=pm.distributions.transforms.ZeroSumTransform([-1]),
        )
        fprobs = pm.math.softmax(s_mu[None, :] + phi[None, :] * mu[:, None], axis=-1)

        pm.Multinomial("y_res", p=fprobs, n=np.ones(len(inp)), observed=res, dims=("obs", "outp"))

    return mdl


@pytest.mark.parametrize("jitter", [12.0, 750.0, 1000.0])
def test_unstable_lbfgs_update_mask(capsys, jitter):
    model = unstable_lbfgs_update_mask_model()

    if jitter < 750.0:
        # Low jitter values should succeed
        with model:
            idata = pmx.fit(
                method="pathfinder",
                jitter=jitter,
                random_seed=4,
            )
        out, err = capsys.readouterr()
        status_pattern = [
            r"INIT_FAILED_LOW_UPDATE_PCT\s+\d+",
            r"LOW_UPDATE_PCT\s+\d+",
            r"LBFGS_FAILED\s+\d+",
            r"SUCCESS\s+\d+",
        ]
        for pattern in status_pattern:
            assert re.search(pattern, out) is not None

    else:
        # High jitter values (>=500) cause numerical overflow and all paths fail
        # jitter=500 raises "All paths failed", jitter=1000 fails earlier with "BUG: Failed to iterate"
        with pytest.raises(ValueError, match="(All paths failed|BUG: Failed to iterate)"):
            with model:
                idata = pmx.fit(
                    method="pathfinder",
                    jitter=jitter,
                    random_seed=4,
                    num_paths=4,
                )


@pytest.mark.parametrize("inference_backend", ["pymc", "blackjax"])
@pytest.mark.filterwarnings("ignore:JAXopt is no longer maintained.:DeprecationWarning")
def test_pathfinder(inference_backend, reference_idata):
    if inference_backend == "blackjax" and sys.platform == "win32":
        pytest.skip("JAX not supported on windows")

    if inference_backend == "blackjax":
        pytest.importorskip("blackjax")
        model = eight_schools_model()
        with model:
            idata = pmx.fit(
                method="pathfinder",
                num_paths=10,
                jitter=12.0,
                random_seed=41,
                inference_backend=inference_backend,
                add_pathfinder_groups=False,  # Diagnostic groups not supported with blackjax
            )
    else:
        idata = reference_idata
        np.testing.assert_allclose(idata.posterior["mu"].mean(), 5.0, atol=0.95)
        np.testing.assert_allclose(idata.posterior["tau"].mean(), 4.15, atol=1.35)

    assert idata.posterior["mu"].shape == (1, 1000)
    assert idata.posterior["tau"].shape == (1, 1000)
    assert idata.posterior["theta"].shape == (1, 1000, 8)


@pytest.mark.parametrize(
    "concurrent",
    [pytest.param("thread", marks=pytest.mark.skip(reason="CI hangs on Windows")), "process"],
)
def test_concurrent_results(reference_idata, concurrent):
    model = eight_schools_model()
    with model:
        idata_conc = pmx.fit(
            method="pathfinder",
            num_paths=10,
            jitter=12.0,
            random_seed=41,
            inference_backend="pymc",
            concurrent=concurrent,
        )

    np.testing.assert_allclose(
        reference_idata.posterior.mu.data.mean(),
        idata_conc.posterior.mu.data.mean(),
        atol=0.4,
    )

    np.testing.assert_allclose(
        reference_idata.posterior.tau.data.mean(),
        idata_conc.posterior.tau.data.mean(),
        atol=0.4,
    )


def test_seed(reference_idata):
    model = eight_schools_model()
    with model:
        idata_41 = pmx.fit(
            method="pathfinder",
            num_paths=4,
            jitter=10.0,
            random_seed=41,
            inference_backend="pymc",
        )

        idata_123 = pmx.fit(
            method="pathfinder",
            num_paths=4,
            jitter=10.0,
            random_seed=123,
            inference_backend="pymc",
        )

    assert not np.allclose(idata_41.posterior.mu.data.mean(), idata_123.posterior.mu.data.mean())

    assert np.allclose(idata_41.posterior.mu.data.mean(), idata_41.posterior.mu.data.mean())


def test_bfgs_sample():
    import pytensor.tensor as pt

    from pymc_extras.inference.pathfinder.pathfinder import (
        alpha_recover,
        bfgs_sample,
        inverse_hessian_factors,
    )

    """test BFGS sampling"""
    Lp1, N = 8, 10
    L = Lp1 - 1
    J = 6
    num_samples = 1000

    # mock data
    x_data = np.random.randn(Lp1, N)
    g_data = np.random.randn(Lp1, N)

    # get factors
    x_full = pt.as_tensor(x_data, dtype="float64")
    g_full = pt.as_tensor(g_data, dtype="float64")

    x = x_full[1:]
    g = g_full[1:]
    alpha, s, z = alpha_recover(x_full, g_full)
    beta, gamma = inverse_hessian_factors(alpha, s, z, J)

    # sample
    phi, logq = bfgs_sample(
        num_samples=num_samples,
        x=x,
        g=g,
        alpha=alpha,
        beta=beta,
        gamma=gamma,
    )

    # check shapes
    assert beta.eval().shape == (L, N, 2 * J)
    assert gamma.eval().shape == (L, 2 * J, 2 * J)
    assert all(phi.shape.eval() == (L, num_samples, N))
    assert all(logq.shape.eval() == (L, num_samples))


@pytest.mark.parametrize("importance_sampling", ["psis", "psir", "identity", None])
def test_pathfinder_importance_sampling(importance_sampling):
    model = eight_schools_model()

    num_paths = 4
    num_draws_per_path = 300
    num_draws = 750

    with model:
        idata = pmx.fit(
            method="pathfinder",
            num_paths=num_paths,
            num_draws_per_path=num_draws_per_path,
            num_draws=num_draws,
            maxiter=5,
            random_seed=41,
            inference_backend="pymc",
            importance_sampling=importance_sampling,
        )

    if importance_sampling is None:
        assert idata.posterior["mu"].shape == (num_paths, num_draws_per_path)
        assert idata.posterior["tau"].shape == (num_paths, num_draws_per_path)
        assert idata.posterior["theta"].shape == (num_paths, num_draws_per_path, 8)
    else:
        assert idata.posterior["mu"].shape == (1, num_draws)
        assert idata.posterior["tau"].shape == (1, num_draws)
        assert idata.posterior["theta"].shape == (1, num_draws, 8)


def test_pathfinder_initvals():
    # Run a model with an ordered transform that will fail unless initvals are in place
    with pm.Model() as mdl:
        pm.Normal("ordered", size=10, transform=pm.distributions.transforms.ordered)
        idata = pmx.fit_pathfinder(initvals={"ordered": np.linspace(0, 1, 10)})

    # Check that the samples are ordered to make sure transform was applied
    assert np.all(
        idata.posterior["ordered"][..., 1:].values > idata.posterior["ordered"][..., :-1].values
    )



#   Copyright 2024 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import numpy as np
import pymc as pm
import pytest

import pymc_extras as pmx

from pymc_extras.inference.laplace_approx.find_map import GradientBackend
from pymc_extras.inference.laplace_approx.laplace import (
    fit_laplace,
    get_conditional_gaussian_approximation,
)

pytestmark = pytest.mark.filterwarnings(
    "ignore:Numba will use object mode to run MinimizeOp:UserWarning"
)


@pytest.fixture(scope="session")
def rng():
    seed = sum(map(ord, "test_laplace"))
    return np.random.default_rng(seed)


@pytest.mark.parametrize("vectorize_draws", (True, False))
@pytest.mark.parametrize(
    "mode, gradient_backend",
    [(None, "pytensor"), ("NUMBA", "pytensor"), pytest.param("JAX", "jax"), ("JAX", "pytensor")],
)
def test_fit_laplace_basic(mode, gradient_backend: GradientBackend, vectorize_draws):
    # Example originates from Bayesian Data Analyses, 3rd Edition
    # By Andrew Gelman, John Carlin, Hal Stern, David Dunson,
    # Aki Vehtari, and Donald Rubin.
    # See section. 4.1

    y = np.array([2642, 3503, 4358], dtype=np.float64)
    n = y.size
    draws = 100000

    with pm.Model() as m:
        mu = pm.Flat("mu")
        logsigma = pm.Flat("logsigma")

        yobs = pm.Normal("y", mu=mu, sigma=pm.math.exp(logsigma), observed=y)
        vars = [mu, logsigma]

        idata = pmx.fit(
            method="laplace",
            optimize_method="trust-ncg",
            draws=draws,
            random_seed=173300,
            compile_kwargs={"mode": mode},
            gradient_backend=gradient_backend,
            optimizer_kwargs=dict(tol=1e-20),
            vectorize_draws=vectorize_draws,
            progressbar=False,
        )

    assert idata.posterior["mu"].shape == (1, draws)
    assert idata.posterior["logsigma"].shape == (1, draws)
    assert idata.observed_data["y"].shape == (n,)
    assert idata.fit["mean_vector"].shape == (len(vars),)
    assert idata.fit["covariance_matrix"].shape == (len(vars), len(vars))

    bda_map = [np.log(y.std()), y.mean()]
    bda_cov = np.array([[1 / (2 * n), 0], [0, y.var() / n]])

    np.testing.assert_allclose(idata.posterior["logsigma"].mean(), bda_map[0], rtol=1e-3)
    np.testing.assert_allclose(idata.posterior["mu"].mean(), bda_map[1], atol=1)

    np.testing.assert_allclose(idata.fit["mean_vector"].values, bda_map, atol=1, rtol=1e-3)
    np.testing.assert_allclose(idata.fit["covariance_matrix"].values, bda_cov, rtol=1e-3, atol=1e-3)


def test_fit_laplace_outside_model_context():
    with pm.Model() as m:
        mu = pm.Normal("mu", 0, 1)
        sigma = pm.Exponential("sigma", 1)
        y_hat = pm.Normal("y_hat", mu=mu, sigma=sigma, observed=np.random.normal(size=10))

    idata = fit_laplace(
        model=m,
        optimize_method="L-BFGS-B",
        use_grad=True,
        progressbar=False,
        vectorize_draws=False,
        draws=100,
    )

    assert hasattr(idata, "posterior")
    assert hasattr(idata, "fit")
    assert hasattr(idata, "optimizer_result")

    assert idata.posterior["mu"].shape == (1, 100)


@pytest.mark.parametrize(
    "include_transformed", [True, False], ids=["include_transformed", "no_transformed"]
)
def test_fit_laplace_coords(include_transformed, rng):
    coords = {"city": ["A", "B", "C"], "obs_idx": np.arange(100)}
    with pm.Model(coords=coords) as model:
        mu = pm.Normal("mu", mu=3, sigma=0.5, dims=["city"])
        sigma = pm.Exponential("sigma", 1, dims=["city"])
        obs = pm.Normal(
            "obs",
            mu=mu,
            sigma=sigma,
            observed=rng.normal(loc=3, scale=1.5, size=(100, 3)),
            dims=["obs_idx", "city"],
        )

        idata = pmx.fit(
            method="laplace",
            optimize_method="trust-ncg",
            draws=1000,
            optimizer_kwargs=dict(tol=1e-20),
            include_transformed=include_transformed,
            progressbar=False,
        )

    np.testing.assert_allclose(
        idata.posterior.mu.mean(dim=["chain", "draw"]).values, np.full((3,), 3), atol=0.5
    )
    np.testing.assert_allclose(
        idata.posterior.sigma.mean(dim=["chain", "draw"]).values, np.full((3,), 1.5), atol=0.3
    )

    assert idata.fit.rows.values.tolist() == [
        "sigma_log__[A]",
        "sigma_log__[B]",
        "sigma_log__[C]",
        "mu[A]",
        "mu[B]",
        "mu[C]",
    ]

    assert hasattr(idata, "unconstrained_posterior") == include_transformed
    if include_transformed:
        assert "sigma_log__" in idata.unconstrained_posterior
        assert "city" in idata.unconstrained_posterior.coords


@pytest.mark.parametrize(
    "draws, use_dims",
    [(500, False), (500, True), (1000, False), (1000, True)],
)
def test_fit_laplace_ragged_coords(draws, use_dims, rng):
    coords = {"city": ["A", "B", "C"], "feature": [0, 1], "obs_idx": np.arange(100)}
    with pm.Model(coords=coords) as ragged_dim_model:
        X = pm.Data("X", np.ones((100, 2)), dims=["obs_idx", "feature"] if use_dims else None)
        beta = pm.Normal(
            "beta", mu=[[-100.0, 100.0], [-100.0, 100.0], [-100.0, 100.0]], dims=["city", "feature"]
        )
        mu = pm.Deterministic(
            "mu",
            (X[:, None, :] * beta[None]).sum(axis=-1),
            dims=["obs_idx", "city"] if use_dims else None,
        )
        sigma = pm.Normal("sigma", mu=1.5, sigma=0.5, dims=["city"])

        obs = pm.Normal(
            "obs",
            mu=mu,
            sigma=sigma,
            observed=rng.normal(loc=3, scale=1.5, size=(100, 3)),
            dims=["obs_idx", "city"],
        )

        idata = fit_laplace(
            optimize_method="Newton-CG",
            progressbar=False,
            use_grad=True,
            use_hessp=True,
            draws=draws,
        )

    # These should have been dropped when the laplace idata was created
    assert "laplace_approximation" not in list(idata.posterior.data_vars.keys())
    assert "unpacked_var_names" not in list(idata.posterior.coords.keys())

    assert idata["posterior"].beta.shape[-2:] == (3, 2)
    assert idata["posterior"].sigma.shape[-1:] == (3,)
    assert idata["posterior"].chain.shape[0] == 1
    assert idata["posterior"].draw.shape[0] == draws

    # Check that everything got unraveled correctly -- feature 0 should be strictly negative, feature 1
    # strictly positive
    assert (idata["posterior"].beta.sel(feature=0).to_numpy() < 0).all()
    assert (idata["posterior"].beta.sel(feature=1).to_numpy() > 0).all()


def test_model_with_nonstandard_dimensionality(rng):
    y_obs = np.concatenate(
        [rng.normal(-1, 2, size=150), rng.normal(3, 1, size=350), rng.normal(5, 4, size=50)]
    )

    with pm.Model(coords={"obs_idx": range(y_obs.size), "class": ["A", "B", "C"]}) as model:
        y = pm.Data("y", y_obs, dims=["obs_idx"])

        mu = pm.Normal("mu", mu=1, sigma=3, dims=["class"])
        sigma = pm.HalfNormal("sigma", sigma=3, dims=["class"])

        w = pm.Dirichlet(
            "w",
            a=np.ones(
                3,
            ),
            dims=["class"],
        )
        class_idx = pm.Categorical("class_idx", p=w, dims=["obs_idx"])
        y_hat = pm.Normal(
            "obs", mu=mu[class_idx], sigma=sigma[class_idx], observed=y, dims=["obs_idx"]
        )

    with pmx.marginalize(model, [class_idx]):
        idata = pmx.fit_laplace(progressbar=False)

    # The dirichlet value variable has a funky shape; check that it got a default
    assert "w_simplex___dim_0" in list(idata.unconstrained_posterior.w_simplex__.coords.keys())
    assert "class" not in list(idata.unconstrained_posterior.w_simplex__.coords.keys())
    assert len(idata.unconstrained_posterior.coords["w_simplex___dim_0"]) == 2

    # On the other hand, check that the actual w has the correct dims
    assert "class" in list(idata.posterior.w.coords.keys())

    # The log transform is 1-to-1, so it should have the same dims as the original rv
    assert "class" in list(idata.unconstrained_posterior.sigma_log__.coords.keys())


def test_laplace_nonstandard_dims_2d():
    true_P = np.array([[0.5, 0.3, 0.2], [0.1, 0.6, 0.3], [0.2, 0.4, 0.4]])
    y_obs = pm.draw(
        pmx.DiscreteMarkovChain.dist(
            P=true_P,
            init_dist=pm.Categorical.dist(
                logit_p=np.ones(
                    3,
                )
            ),
            shape=(100, 5),
        )
    )

    with pm.Model(
        coords={
            "time": range(y_obs.shape[0]),
            "state": list("ABC"),
            "next_state": list("ABC"),
            "unit": [1, 2, 3, 4, 5],
        }
    ) as model:
        y = pm.Data("y", y_obs, dims=["time", "unit"])
        init_dist = pm.Categorical.dist(
            logit_p=np.ones(
                3,
            )
        )
        P = pm.Dirichlet("P", a=np.eye(3) * 2 + 1, dims=["state", "next_state"])
        y_hat = pmx.DiscreteMarkovChain(
            "y_hat", P=P, init_dist=init_dist, dims=["time", "unit"], observed=y_obs
        )

        idata = pmx.fit_laplace(progressbar=False)

        # The simplex transform should drop from the right-most dimension, so the left dimension should be unmodified
        assert "state" in list(idata.unconstrained_posterior.P_simplex__.coords.keys())

        # The mutated dimension should be unknown coords
        assert "P_simplex___dim_1" in list(idata.unconstrained_posterior.P_simplex__.coords.keys())

        assert idata.unconstrained_posterior.P_simplex__.shape[-2:] == (3, 2)


def test_laplace_nonscalar_rv_without_dims():
    with pm.Model(coords={"test": ["A", "B", "C"]}) as model:
        x_loc = pm.Normal("x_loc", mu=0, sigma=1, dims=["test"])
        x = pm.Normal("x", mu=x_loc, sigma=1, shape=(2, 3))
        y = pm.Normal("y", mu=x, sigma=1, observed=np.random.randn(10, 2, 3))

        idata = pmx.fit_laplace(progressbar=False)

    assert idata.posterior["x"].shape == (1, 500, 2, 3)
    assert all(f"x_dim_{i}" in idata.posterior.coords for i in range(2))
    assert idata.fit.rows.values.tolist() == [
        "x_loc[A]",
        "x_loc[B]",
        "x_loc[C]",
        "x[0,0]",
        "x[0,1]",
        "x[0,2]",
        "x[1,0]",
        "x[1,1]",
        "x[1,2]",
    ]


# Test these three optimizers because they are either special cases for H_inv (BFGS, L-BFGS-B) or are
# gradient free and require re-compilation of hessp (powell).
@pytest.mark.parametrize("optimizer_method", ["BFGS", "L-BFGS-B", "powell"])
def test_laplace_scalar_basinhopping(optimizer_method):
    # Example model from Statistical Rethinking
    data = np.array([0, 0, 0, 1, 1, 1, 1, 1, 1])

    with pm.Model():
        p = pm.Uniform("p", 0, 1)
        w = pm.Binomial("w", n=len(data), p=p, observed=data.sum())

        idata_laplace = pmx.fit_laplace(
            optimize_method="basinhopping",
            optimizer_kwargs={"minimizer_kwargs": {"method": optimizer_method}, "niter": 1},
            progressbar=False,
        )

    assert idata_laplace.fit.mean_vector.shape == (1,)
    assert idata_laplace.fit.covariance_matrix.shape == (1, 1)

    np.testing.assert_allclose(
        idata_laplace.posterior.p.mean(dim=["chain", "draw"]), data.mean(), atol=0.1
    )


def test_get_conditional_gaussian_approximation():
    """
    Consider the trivial case of:

    y | x ~ N(x, cov_param)
    x | param ~ N(mu_param, Q^-1)

    cov_param ~ N(cov_mu, cov_cov)
    mu_param ~ N(mu_mu, mu_cov)
    Q ~ N(Q_mu, Q_cov)

    This has an analytic solution at the mode which we can compare against.
    """
    rng = np.random.default_rng(12345)
    n = 10000
    d = 10

    # Initialise arrays
    mu_true = rng.random(d)
    cov_true = np.diag(rng.random(d))
    Q_val = np.diag(rng.random(d))
    cov_param_val = np.diag(rng.random(d))

    x_val = rng.random(d)
    mu_val = rng.random(d)

    mu_mu = rng.random(d)
    mu_cov = np.diag(np.ones(d))
    cov_mu = rng.random(d**2)
    cov_cov = np.diag(np.ones(d**2))
    Q_mu = rng.random(d**2)
    Q_cov = np.diag(np.ones(d**2))

    with pm.Model() as model:
        y_obs = rng.multivariate_normal(mean=mu_true, cov=cov_true, size=n)

        mu_param = pm.MvNormal("mu_param", mu=mu_mu, cov=mu_cov)
        cov_param = pm.MvNormal("cov_param", mu=cov_mu, cov=cov_cov)
        Q = pm.MvNormal("Q", mu=Q_mu, cov=Q_cov)

        # Pytensor currently doesn't support autograd for pt inverses, so we use a numeric Q instead
        x = pm.MvNormal("x", mu=mu_param, cov=np.linalg.inv(Q_val))

        y = pm.MvNormal(
            "y",
            mu=x,
            cov=cov_param.reshape((d, d)),
            observed=y_obs,
        )

        # logp(x | y, params)
        cga = get_conditional_gaussian_approximation(
            x=model.rvs_to_values[x],
            Q=Q.reshape((d, d)),
            mu=mu_param,
            optimizer_kwargs={"tol": 1e-25},
        )

    x0, log_x_posterior = cga(
        x=x_val, mu_param=mu_val, cov_param=cov_param_val.flatten(), Q=Q_val.flatten()
    )

    # Get analytic values of the mode and Laplace-approximated log posterior
    cov_param_inv = np.linalg.inv(cov_param_val)

    x0_true = np.linalg.inv(n * cov_param_inv + 2 * Q_val) @ (
        cov_param_inv @ y_obs.sum(axis=0) + 2 * Q_val @ mu_val
    )

    jac_true = cov_param_inv @ (y_obs - x0_true).sum(axis=0) - Q_val @ (x0_true - mu_val)
    hess_true = -n * cov_param_inv - Q_val

    log_x_posterior_laplace_true = (
        -0.5 * x_val.T @ (-hess_true + Q_val) @ x_val
        + x_val.T @ (Q_val @ mu_val + jac_true - hess_true @ x0_true)
        + 0.5 * np.log(np.linalg.det(Q_val))
    )

    np.testing.assert_allclose(x0, x0_true, atol=0.1, rtol=0.1)
    np.testing.assert_allclose(log_x_posterior, log_x_posterior_laplace_true, atol=0.1, rtol=0.1)



import logging

from collections.abc import Callable
from typing import Literal, cast

import numpy as np
import pymc as pm

from better_optimize import basinhopping, minimize
from better_optimize.constants import minimize_method
from pymc.blocking import DictToArrayBijection, RaveledVars
from pymc.initial_point import make_initial_point_fn
from pymc.model.transform.optimization import freeze_dims_and_data
from pymc.util import get_default_varnames
from pytensor.tensor import TensorVariable
from scipy.optimize import OptimizeResult

from pymc_extras.inference.laplace_approx.idata import (
    add_data_to_inference_data,
    add_fit_to_inference_data,
    add_optimizer_result_to_inference_data,
    map_results_to_inference_data,
)
from pymc_extras.inference.laplace_approx.scipy_interface import (
    GradientBackend,
    scipy_optimize_funcs_from_loss,
    set_optimizer_function_defaults,
)

_log = logging.getLogger(__name__)


def get_nearest_psd(A: np.ndarray) -> np.ndarray:
    """
    Compute the nearest positive semi-definite matrix to a given matrix.

    This function takes a square matrix and returns the nearest positive semi-definite matrix using
    eigenvalue decomposition. It ensures all eigenvalues are non-negative. The "nearest" matrix is defined in terms
    of the Frobenius norm.

    Parameters
    ----------
    A : np.ndarray
        Input square matrix.

    Returns
    -------
    np.ndarray
        The nearest positive semi-definite matrix to the input matrix.
    """
    C = (A + A.T) / 2
    eigval, eigvec = np.linalg.eigh(C)
    eigval[eigval < 0] = 0

    return eigvec @ np.diag(eigval) @ eigvec.T


def _make_initial_point(model, initvals=None, random_seed=None, jitter_rvs=None):
    jitter_rvs = [] if jitter_rvs is None else jitter_rvs

    ipfn = make_initial_point_fn(
        model=model,
        jitter_rvs=set(jitter_rvs),
        return_transformed=True,
        overrides=initvals,
    )

    start_dict = ipfn(random_seed)
    vars_dict = {var.name: var for var in model.continuous_value_vars}
    initial_params = DictToArrayBijection.map(
        {var_name: value for var_name, value in start_dict.items() if var_name in vars_dict}
    )

    return initial_params


def _compute_inverse_hessian(
    optimizer_result: OptimizeResult | None,
    optimal_point: np.ndarray | None,
    f_fused: Callable | None,
    f_hessp: Callable | None,
    use_hess: bool,
    method: minimize_method | Literal["BFGS", "L-BFGS-B"],
):
    """
    Compute the Hessian matrix or its inverse based on the optimization result and the method used.

    Downstream functions (e.g. laplace approximation) will need the inverse Hessian matrix. This function computes it
    in the cheapest way possible, depending on the optimization method used and the available compiled functions.

    Parameters
    ----------
    optimizer_result: OptimizeResult, optional
        The result of the optimization, containing the optimized parameters and possibly an approximate inverse Hessian.
    optimal_point: np.ndarray, optional
        The optimal point found by the optimizer, used to compute the Hessian if necessary. If not provided, it will be
        extracted from the optimizer result.
    f_fused: callable, optional
        The compiled function representing the loss and possibly its gradient and Hessian.
    f_hessp: callable, optional
        The compiled function for Hessian-vector products, if available.
    use_hess: bool
        Whether the Hessian matrix was used in the optimization.
    method: minimize_method
        The optimization method used, which determines how the Hessian is computed.

    Returns
    -------
    H_inv: np.ndarray
        The inverse Hessian matrix, computed based on the optimization method and available functions.
    """
    if optimal_point is None and optimizer_result is None:
        raise ValueError("At least one of `optimal_point` or `optimizer_result` must be provided.")

    x_star = optimizer_result.x if optimizer_result is not None else optimal_point
    n_vars = len(x_star)

    if method == "BFGS" and optimizer_result is not None:
        # If we used BFGS, the optimizer result will contain the inverse Hessian -- we can just use that rather than
        # re-computing something
        if hasattr(optimizer_result, "lowest_optimization_result"):
            # We did basinhopping, need to get the inner optimizer results
            H_inv = getattr(optimizer_result.lowest_optimization_result, "hess_inv", None)
        else:
            H_inv = getattr(optimizer_result, "hess_inv", None)

    elif method == "L-BFGS-B" and optimizer_result is not None:
        # Here we will have a LinearOperator representing the inverse Hessian-Vector product.
        if hasattr(optimizer_result, "lowest_optimization_result"):
            # We did basinhopping, need to get the inner optimizer results
            f_hessp_inv = getattr(optimizer_result.lowest_optimization_result, "hess_inv", None)
        else:
            f_hessp_inv = getattr(optimizer_result, "hess_inv", None)

        if f_hessp_inv is not None:
            basis = np.eye(n_vars)
            H_inv = np.stack([f_hessp_inv(basis[:, i]) for i in range(n_vars)], axis=-1)
        else:
            H_inv = None

    elif f_hessp is not None:
        # In the case that hessp was used, the results object will not save the inverse Hessian, so we can compute it from
        # the hessp function, using euclidian basis vector.
        basis = np.eye(n_vars)
        H = np.stack([f_hessp(x_star, basis[:, i]) for i in range(n_vars)], axis=-1)
        H_inv = np.linalg.inv(get_nearest_psd(H))

    elif use_hess and f_fused is not None:
        # If we compiled a hessian function, just use it
        _, _, H = f_fused(x_star)
        H_inv = np.linalg.inv(get_nearest_psd(H))

    else:
        H_inv = None

    return H_inv


def find_MAP(
    method: minimize_method | Literal["basinhopping"] = "L-BFGS-B",
    *,
    model: pm.Model | None = None,
    use_grad: bool | None = None,
    use_hessp: bool | None = None,
    use_hess: bool | None = None,
    initvals: dict | None = None,
    random_seed: int | np.random.Generator | None = None,
    jitter_rvs: list[TensorVariable] | None = None,
    progressbar: bool = True,
    include_transformed: bool = True,
    freeze_model: bool = True,
    gradient_backend: GradientBackend = "pytensor",
    compile_kwargs: dict | None = None,
    compute_hessian: bool = False,
    **optimizer_kwargs,
) -> (
    dict[str, np.ndarray]
    | tuple[dict[str, np.ndarray], np.ndarray]
    | tuple[dict[str, np.ndarray], OptimizeResult]
    | tuple[dict[str, np.ndarray], OptimizeResult, np.ndarray]
):
    """
    Fit a PyMC model via maximum a posteriori (MAP) estimation using JAX and scipy.optimize.

    Parameters
    ----------
    model : pm.Model
        The PyMC model to be fit. If None, the current model context is used.
    method : str
        The optimization method to use. Valid choices are: Nelder-Mead, Powell, CG, BFGS, L-BFGS-B, TNC, SLSQP,
        trust-constr, dogleg, trust-ncg, trust-exact, trust-krylov, and basinhopping.

        See scipy.optimize.minimize documentation for details.
    use_grad : bool | None, optional
        Whether to use gradients in the optimization. Defaults to None, which determines this automatically based on
        the ``method``.
    use_hessp : bool | None, optional
        Whether to use Hessian-vector products in the optimization. Defaults to None, which determines this automatically based on
        the ``method``.
    use_hess : bool | None, optional
        Whether to use the Hessian matrix in the optimization. Defaults to None, which determines this automatically based on
        the ``method``.
    initvals : None | dict, optional
        Initial values for the model parameters, as str:ndarray key-value pairs. Partial initialization is permitted.
         If None, the model's default initial values are used.
    random_seed : None | int | np.random.Generator, optional
        Seed for the random number generator or a numpy Generator for reproducibility
    jitter_rvs : list of TensorVariables, optional
        Variables whose initial values should be jittered. If None, all variables are jittered.
    progressbar : bool, optional
        Whether to display a progress bar during optimization. Defaults to True.
    include_transformed: bool, optional
        Whether to include transformed variable values in the returned dictionary. Defaults to True.
    freeze_model: bool, optional
        If True, freeze_dims_and_data will be called on the model before compiling the loss functions. This is
        sometimes necessary for JAX, and can sometimes improve performance by allowing constant folding. Defaults to
        True.
    gradient_backend: str, default "pytensor"
        Which backend to use to compute gradients. Must be one of "pytensor" or "jax".
    compute_hessian: bool
        If True, the inverse Hessian matrix at the optimum will be computed and included in the returned
        InferenceData object. This is needed for the Laplace approximation, but can be computationally expensive for
        high-dimensional problems. Defaults to False.
    compile_kwargs: dict, optional
        Additional options to pass to the ``pytensor.function`` function when compiling loss functions.
    **optimizer_kwargs
        Additional keyword arguments to pass to the ``scipy.optimize`` function being used. Unless
        ``method = "basinhopping"``, ``scipy.optimize.minimize`` will be used. For ``basinhopping``,
        ``scipy.optimize.basinhopping`` will be used. See the documentation of these functions for details.

    Returns
    -------
    map_result: az.InferenceData
        Results of Maximum A Posteriori (MAP) estimation, including the optimized point, inverse Hessian, transformed
        latent variables, and optimizer results.
    """
    compile_kwargs = {} if compile_kwargs is None else compile_kwargs
    model = pm.modelcontext(model) if model is None else model

    if freeze_model:
        model = freeze_dims_and_data(model)

    initial_params = _make_initial_point(model, initvals, random_seed, jitter_rvs)

    do_basinhopping = method == "basinhopping"
    minimizer_kwargs = optimizer_kwargs.pop("minimizer_kwargs", {})

    if do_basinhopping:
        # For a nice API, we let the user set method="basinhopping", but if we're doing basinhopping we still need
        # another method for the inner optimizer. This will be set in the minimizer_kwargs, but also needs a default
        # if one isn't provided.

        method = minimizer_kwargs.pop("method", "L-BFGS-B")
        minimizer_kwargs["method"] = method

    use_grad, use_hess, use_hessp = set_optimizer_function_defaults(
        method, use_grad, use_hess, use_hessp
    )

    f_fused, f_hessp = scipy_optimize_funcs_from_loss(
        loss=-model.logp(jacobian=False),
        inputs=model.continuous_value_vars + model.discrete_value_vars,
        initial_point_dict=DictToArrayBijection.rmap(initial_params),
        use_grad=use_grad,
        use_hess=use_hess,
        use_hessp=use_hessp,
        gradient_backend=gradient_backend,
        compile_kwargs=compile_kwargs,
    )

    args = optimizer_kwargs.pop("args", ())

    # better_optimize.minimize will check if f_logp is a fused loss+grad Op, and automatically assign the jac argument
    # if so. That is why the jac argument is not passed here in either branch.

    if do_basinhopping:
        if "args" not in minimizer_kwargs:
            minimizer_kwargs["args"] = args
        if "hessp" not in minimizer_kwargs:
            minimizer_kwargs["hessp"] = f_hessp
        if "method" not in minimizer_kwargs:
            minimizer_kwargs["method"] = method

        optimizer_result = basinhopping(
            func=f_fused,
            x0=cast(np.ndarray[float], initial_params.data),
            progressbar=progressbar,
            minimizer_kwargs=minimizer_kwargs,
            **optimizer_kwargs,
        )

    else:
        optimizer_result = minimize(
            f=f_fused,
            x0=cast(np.ndarray[float], initial_params.data),
            args=args,
            hessp=f_hessp,
            progressbar=progressbar,
            method=method,
            **optimizer_kwargs,
        )

    if compute_hessian:
        H_inv = _compute_inverse_hessian(
            optimizer_result=optimizer_result,
            optimal_point=None,
            f_fused=f_fused,
            f_hessp=f_hessp,
            use_hess=use_hess,
            method=method,
        )
    else:
        H_inv = None

    raveled_optimized = RaveledVars(optimizer_result.x, initial_params.point_map_info)
    unobserved_vars = get_default_varnames(model.unobserved_value_vars, include_transformed=True)
    unobserved_vars_values = model.compile_fn(unobserved_vars, mode="FAST_COMPILE")(
        DictToArrayBijection.rmap(raveled_optimized)
    )

    optimized_point = {
        var.name: value for var, value in zip(unobserved_vars, unobserved_vars_values)
    }

    idata = map_results_to_inference_data(
        map_point=optimized_point, model=model, include_transformed=include_transformed
    )

    idata = add_fit_to_inference_data(idata=idata, mu=raveled_optimized, H_inv=H_inv, model=model)

    idata = add_optimizer_result_to_inference_data(
        idata=idata, result=optimizer_result, method=method, mu=raveled_optimized, model=model
    )

    idata = add_data_to_inference_data(
        idata=idata, progressbar=False, model=model, compile_kwargs=compile_kwargs
    )

    return idata



#   Copyright 2025 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytensor.tensor as pt

from pymc.logprob.transforms import Transform

__all__ = ["PartialOrder"]


def dtype_minval(dtype):
    """Find the minimum value for a given dtype"""
    return np.iinfo(dtype).min if np.issubdtype(dtype, np.integer) else np.finfo(dtype).min


def padded_where(x, to_len, padval=-1):
    """A padded version of np.where"""
    w = np.where(x)
    return np.concatenate([w[0], np.full(to_len - len(w[0]), padval)])


class PartialOrder(Transform):
    """Create a PartialOrder transform

    A more flexible version of the pymc ordered transform that
    allows specifying a (strict) partial order on the elements.

    Examples
    --------
    .. code:: python

        import numpy as np
        import pymc as pm
        import pymc_extras as pmx

        # Define two partial orders on 4 elements
        # am[i,j] = 1 means i < j
        adj_mats = np.array([
            # 0 < {1, 2} < 3
            [[0, 1, 1, 0],
            [0, 0, 0, 1],
            [0, 0, 0, 1],
            [0, 0, 0, 0]],

            # 1 < 0 < 3 < 2
            [[0, 0, 0, 1],
            [1, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 1, 0]],
        ])

        # Create the partial order from the adjacency matrices
        po = pmx.PartialOrder(adj_mats)

        with pm.Model() as model:
            # Generate 3 samples from both partial orders
            pm.Normal("po_vals", shape=(3,2,4), transform=po,
                        initval=po.initvals((3,2,4)))

            idata = pm.sample()

        # Verify that for first po, the zeroth element is always the smallest
        assert (idata.posterior['po_vals'][:,:,:,0,0] <
                idata.posterior['po_vals'][:,:,:,0,1:]).all()

        # Verify that for second po, the second element is always the largest
        assert (idata.posterior['po_vals'][:,:,:,1,2] >=
                idata.posterior['po_vals'][:,:,:,1,:]).all()

    Technical notes
    ----------------
    Partial order needs to be strict, i.e. without equalities.
    A DAG defining the partial order is sufficient, as transitive closure is automatically computed.
    Code works in O(N*D) in runtime, but takes O(N^3) in initialization,
    where N is the number of nodes in the dag and D is the maximum
    in-degree of a node in the transitive reduction.
    """

    name = "partial_order"

    def __init__(self, adj_mat):
        """
        Initialize the PartialOrder transform

        Parameters
        ----------
        adj_mat: ndarray
            adjacency matrix for the DAG that generates the partial order,
            where ``adj_mat[i][j] = 1`` denotes ``i < j``.
            Note this also accepts multiple DAGs if RV is multidimensional
        """

        # Basic input checks
        if adj_mat.ndim < 2:
            raise ValueError("Adjacency matrix must have at least 2 dimensions")
        if adj_mat.shape[-2] != adj_mat.shape[-1]:
            raise ValueError("Adjacency matrix is not square")
        if adj_mat.min() != 0 or adj_mat.max() != 1:
            raise ValueError("Adjacency matrix must contain only 0s and 1s")

        # Create index over the first ellipsis dimensions
        idx = np.ix_(*[np.arange(s) for s in adj_mat.shape[:-2]])

        # Transitive closure using Floyd-Warshall
        tc = adj_mat.astype(bool)
        for k in range(tc.shape[-1]):
            tc |= np.logical_and(tc[..., :, k, None], tc[..., None, k, :])

        # Check if the dag is acyclic
        if np.any(tc.diagonal(axis1=-2, axis2=-1)):
            raise ValueError("Partial order contains equalities")

        # Transitive reduction using the closure
        # This gives the minimum description of the partial order
        # This is to minmax the input degree
        adj_mat = tc * (1 - np.matmul(tc, tc))

        # Find the maximum in-degree of the reduced dag
        dag_idim = adj_mat.sum(axis=-2).max()

        # Topological sort
        ts_inds = np.zeros(adj_mat.shape[:-1], dtype=int)
        dm = adj_mat.copy()
        for i in range(adj_mat.shape[1]):
            assert dm.sum(axis=-2).min() == 0  # DAG is acyclic
            nind = np.argmin(dm.sum(axis=-2), axis=-1)
            dm[(*idx, slice(None), nind)] = 1  # Make nind not show up again
            dm[(*idx, nind, slice(None))] = 0  # Allow it's children to show
            ts_inds[(*idx, i)] = nind
        self.ts_inds = ts_inds

        # Change the dag to adjacency lists (with -1 for NA)
        dag_T = np.apply_along_axis(padded_where, axis=-2, arr=adj_mat, padval=-1, to_len=dag_idim)
        self.dag = np.swapaxes(dag_T, -2, -1)
        self.is_start = np.all(self.dag[..., :, :] == -1, axis=-1)

    def initvals(self, shape=None, lower=-1, upper=1):
        """
        Create a set of appropriate initial values for the variable.
        NB! It is important that proper initial values are used,
        as only properly ordered values are in the range of the transform.

        Parameters
        ----------
        shape: tuple, default None
            shape of the initial values. If None, adj_mat[:-1] is used
        lower: float, default -1
            lower bound for the initial values
        upper: float, default 1
            upper bound for the initial values

        Returns
        -------
        vals: ndarray
            initial values for the transformed variable
        """

        if shape is None:
            shape = self.dag.shape[:-1]

        if shape[-len(self.dag.shape[:-1]) :] != self.dag.shape[:-1]:
            raise ValueError("Shape must match the shape of the adjacency matrix")

        # Create the initial values
        vals = np.linspace(lower, upper, self.dag.shape[-2])
        inds = np.argsort(self.ts_inds, axis=-1)
        ivals = vals[inds]

        # Expand the initial values to the extra dimensions
        extra_dims = shape[: -len(self.dag.shape[:-1])]
        ivals = np.tile(ivals, extra_dims + tuple([1] * len(self.dag.shape[:-1])))

        return ivals

    def backward(self, value, *inputs):
        minv = dtype_minval(value.dtype)
        x = pt.concatenate(
            [pt.zeros_like(value), pt.full(value.shape[:-1], minv)[..., None]], axis=-1
        )

        # Indices to allow broadcasting the max over the last dimension
        idx = np.ix_(*[np.arange(s) for s in self.dag.shape[:-2]])
        idx2 = tuple(np.tile(i[:, None], self.dag.shape[-1]) for i in idx)

        # Has to be done stepwise as next steps depend on previous values
        # Also has to be done in topological order, hence the ts_inds
        for i in range(self.dag.shape[-2]):
            tsi = self.ts_inds[..., i]
            if len(tsi.shape) == 0:
                tsi = int(tsi)  # if shape 0, it's a scalar
            ni = (*idx, tsi)  # i-th node in topological order
            eni = (Ellipsis, *ni)
            ist = self.is_start[ni]

            mval = pt.max(x[(Ellipsis, *idx2, self.dag[ni])], axis=-1)
            x = pt.set_subtensor(x[eni], ist * value[eni] + (1 - ist) * (mval + pt.exp(value[eni])))
        return x[..., :-1]

    def forward(self, value, *inputs):
        y = pt.zeros_like(value)

        minv = dtype_minval(value.dtype)
        vx = pt.concatenate([value, pt.full(value.shape[:-1], minv)[..., None]], axis=-1)

        # Indices to allow broadcasting the max over the last dimension
        idx = np.ix_(*[np.arange(s) for s in self.dag.shape[:-2]])
        idx = tuple(np.tile(i[:, None, None], self.dag.shape[-2:]) for i in idx)

        y = self.is_start * value + (1 - self.is_start) * (
            pt.log(value - pt.max(vx[(Ellipsis, *idx, self.dag[..., :])], axis=-1))
        )

        return y

    def log_jac_det(self, value, *inputs):
        return pt.sum(value * (1 - self.is_start), axis=-1)



#   Copyright 2025 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pymc as pm

from pymc_extras.distributions.transforms import PartialOrder


class TestPartialOrder:
    adj_mats = np.array(
        [
            # 0 < {1, 2} < 3
            [[0, 1, 1, 0], [0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 0]],
            # 1 < 0 < 3 < 2
            [[0, 0, 0, 1], [1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 1, 0]],
        ]
    )

    valid_values = np.array([[0, 2, 1, 3], [1, 0, 3, 2]], dtype=float)

    # Test that forward and backward are inverses of eachother
    # And that it works when extra dimensions are added in data
    def test_forward_backward_dimensionality(self):
        po = PartialOrder(self.adj_mats)
        po0 = PartialOrder(self.adj_mats[0])
        vv = self.valid_values
        vv0 = self.valid_values[0]

        testsets = [
            (vv, po),
            (po.initvals(), po),
            (vv0, po0),
            (po0.initvals(), po0),
            (np.tile(vv0, (2, 1)), po0),
            (np.tile(vv0, (2, 3, 2, 1)), po0),
            (np.tile(vv, (2, 3, 2, 1, 1)), po),
        ]

        for vv, po in testsets:
            fw = po.forward(vv)
            bw = po.backward(fw)
            np.testing.assert_allclose(bw.eval(), vv)

    def test_sample_model(self):
        po = PartialOrder(self.adj_mats)
        with pm.Model() as model:
            x = pm.Normal(
                "x",
                size=(3, 2, 4),
                transform=po,
                initval=po.initvals(shape=(3, 2, 4), lower=-1, upper=1),
            )
            idata = pm.sample()

        # Check that the order constraints are satisfied
        # Move chain, draw and "3" dimensions to the back
        xvs = idata.posterior.x.values.transpose(3, 4, 0, 1, 2)
        x0 = xvs[0]  # 0 < {1, 2} < 3
        assert (
            (x0[0] < x0[1]).all()
            and (x0[0] < x0[2]).all()
            and (x0[1] < x0[3]).all()
            and (x0[2] < x0[3]).all()
        )
        x1 = xvs[1]  # 1 < 0 < 3 < 2
        assert (x1[1] < x1[0]).all() and (x1[0] < x1[3]).all() and (x1[3] < x1[2]).all()
